# Test methods with long descriptive names can omit docstrings
# pylint: disable=missing-docstring

import unittest

import numpy as np

from Orange.data import DiscreteVariable, ContinuousVariable, Domain
from Orange.data import Table
from Orange.classification import LogisticRegressionLearner, SklTreeLearner, NaiveBayesLearner,\
                                  MajorityLearner
from Orange.evaluation import AUC, CA, Results, Recall, \
    Precision, TestOnTrainingData, scoring, LogLoss, F1, CrossValidation
from Orange.evaluation.scoring import Specificity
from Orange.preprocess import discretize, Discretize
from Orange.tests import test_filename


class TestScoreMetaType(unittest.TestCase):
    class BaseScore(metaclass=scoring.ScoreMetaType):
        pass

    class Score1(BaseScore, abstract=True):
        class_types = (DiscreteVariable,)

    class Score2(Score1):
        pass

    class Score3(Score2):
        name = "foo"

    class Score4(Score2):
        pass

    class Score5(BaseScore):
        class_types = (DiscreteVariable, ContinuousVariable)

    def test_registry(self):
        """All non-abstract classes appear in the registry"""
        self.assertEqual(
            self.BaseScore.registry,
            {"Score2": self.Score2, "Score3": self.Score3,
             "Score4": self.Score4, "Score5": self.Score5})

    def test_names(self):
        """Attribute `name` defaults to class and is not inherited"""
        self.assertEqual(self.Score2.name, "Score2")
        self.assertEqual(self.Score3.name, "foo")
        self.assertEqual(self.Score4.name, "Score4")
        self.assertEqual(self.Score5.name, "Score5")


class TestPrecision(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.iris = Table('iris')
        cls.score = Precision()

    def test_precision_iris(self):
        learner = LogisticRegressionLearner(preprocessors=[])
        res = TestOnTrainingData()(self.iris, [learner])
        self.assertGreater(self.score(res, average='weighted')[0], 0.95)
        self.assertGreater(self.score(res, target=1)[0], 0.95)
        self.assertGreater(self.score(res, target=1, average=None)[0], 0.95)
        self.assertGreater(self.score(res, target=1, average='weighted')[0], 0.95)
        self.assertGreater(self.score(res, target=0, average=None)[0], 0.99)
        self.assertGreater(self.score(res, target=2, average=None)[0], 0.94)

    def test_precision_multiclass(self):
        results = Results(
            domain=Domain([], DiscreteVariable(name="y", values="01234")),
            actual=[0, 4, 4, 1, 2, 0, 1, 2, 3, 2])
        results.predicted = np.array([[0, 4, 4, 1, 2, 0, 1, 2, 3, 2],
                                      [0, 1, 4, 1, 1, 0, 0, 2, 3, 1]])
        res = self.score(results, average='weighted')
        self.assertEqual(res[0], 1.)
        self.assertAlmostEqual(res[1], 0.78333, 5)

        for target, prob in ((0, 2 / 3),
                             (1, 1 / 4),
                             (2, 1 / 1),
                             (3, 1 / 1),
                             (4, 1 / 1)):
            res = self.score(results, target=target, average=None)
            self.assertEqual(res[0], 1.)
            self.assertEqual(res[1], prob)

    def test_precision_binary(self):
        results = Results(
            domain=Domain([], DiscreteVariable(name="y", values="01")),
            actual=[0, 1, 1, 1, 0, 0, 1, 0, 0, 1])
        results.predicted = np.array([[0, 1, 1, 1, 0, 0, 1, 0, 0, 1],
                                      [0, 1, 1, 1, 0, 0, 1, 1, 1, 0]])
        res = self.score(results)
        self.assertEqual(res[0], 1.)
        self.assertAlmostEqual(res[1], 4 / 6)
        res_target = self.score(results, target=1)
        self.assertEqual(res[0], res_target[0])
        self.assertEqual(res[1], res_target[1])
        res_target = self.score(results, target=0)
        self.assertEqual(res_target[0], 1.)
        self.assertAlmostEqual(res_target[1], 3 / 4)
        res_target = self.score(results, average='macro')
        self.assertEqual(res_target[0], 1.)
        self.assertAlmostEqual(res_target[1], (4 / 6 + 3 / 4) / 2)


class TestRecall(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.iris = Table('iris')
        cls.score = Recall()

    def test_recall_iris(self):
        learner = LogisticRegressionLearner(preprocessors=[])
        res = TestOnTrainingData()(self.iris, [learner])
        self.assertGreater(self.score(res, average='weighted')[0], 0.96)
        self.assertGreater(self.score(res, target=1)[0], 0.9)
        self.assertGreater(self.score(res, target=1, average=None)[0], 0.9)
        self.assertGreater(self.score(res, target=1, average='weighted')[0], 0.9)
        self.assertGreater(self.score(res, target=0, average=None)[0], 0.99)
        self.assertGreater(self.score(res, target=2, average=None)[0], 0.97)

    def test_recall_multiclass(self):
        results = Results(
            domain=Domain([], DiscreteVariable(name="y", values="01234")),
            actual=[0, 4, 4, 1, 2, 0, 1, 2, 3, 2])
        results.predicted = np.array([[0, 4, 4, 1, 2, 0, 1, 2, 3, 2],
                                      [0, 1, 4, 1, 1, 0, 0, 2, 3, 1]])
        res = self.score(results, average='weighted')
        self.assertEqual(res[0], 1.)
        self.assertAlmostEqual(res[1], 0.6)

        for target, prob in ((0, 2 / 2),
                             (1, 1 / 2),
                             (2, 1 / 3),
                             (3, 1 / 1),
                             (4, 1 / 2)):
            res = self.score(results, target=target)
            self.assertEqual(res[0], 1.)
            self.assertEqual(res[1], prob)

    def test_recall_binary(self):
        results = Results(
            domain=Domain([], DiscreteVariable(name="y", values="01")),
            actual=[0, 1, 1, 1, 0, 0, 1, 0, 0, 1])
        results.predicted = np.array([[0, 1, 1, 1, 0, 0, 1, 0, 0, 1],
                                      [0, 1, 1, 1, 0, 0, 1, 1, 1, 0]])
        res = self.score(results)
        self.assertEqual(res[0], 1.)
        self.assertAlmostEqual(res[1], 4 / 5)
        res_target = self.score(results, target=1)
        self.assertEqual(res[0], res_target[0])
        self.assertEqual(res[1], res_target[1])
        res_target = self.score(results, target=0)
        self.assertEqual(res_target[0], 1.)
        self.assertAlmostEqual(res_target[1], 3 / 5)
        res_target = self.score(results, average='macro')
        self.assertEqual(res_target[0], 1.)
        self.assertAlmostEqual(res_target[1], (4 / 5 + 3 / 5) / 2)


class TestF1(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.iris = Table('iris')
        cls.score = F1()

    def test_recall_iris(self):
        learner = LogisticRegressionLearner(preprocessors=[])
        res = TestOnTrainingData()(self.iris, [learner])
        self.assertGreater(self.score(res, average='weighted')[0], 0.95)
        self.assertGreater(self.score(res, target=1)[0], 0.95)
        self.assertGreater(self.score(res, target=1, average=None)[0], 0.95)
        self.assertGreater(self.score(res, target=1, average='weighted')[0], 0.95)
        self.assertGreater(self.score(res, target=0, average=None)[0], 0.99)
        self.assertGreater(self.score(res, target=2, average=None)[0], 0.95)

    def test_F1_multiclass(self):
        results = Results(
            domain=Domain([], DiscreteVariable(name="y", values="01234")),
            actual=[0, 4, 4, 1, 2, 0, 1, 2, 3, 2])
        results.predicted = np.array([[0, 4, 4, 1, 2, 0, 1, 2, 3, 2],
                                      [0, 1, 4, 1, 1, 0, 0, 2, 3, 1]])
        res = self.score(results, average='weighted')
        self.assertEqual(res[0], 1.)
        self.assertAlmostEqual(res[1], 0.61)

        for target, prob in ((0, 4 / 5),
                             (1, 1 / 3),
                             (2, 1 / 2),
                             (3, 1.),
                             (4, 2 / 3)):
            res = self.score(results, target=target)
            self.assertEqual(res[0], 1.)
            self.assertEqual(res[1], prob)

    def test_F1_binary(self):
        results = Results(
            domain=Domain([], DiscreteVariable(name="y", values="01")),
            actual=[0, 1, 1, 1, 0, 0, 1, 0, 0, 1])
        results.predicted = np.array([[0, 1, 1, 1, 0, 0, 1, 0, 0, 1],
                                      [0, 1, 1, 1, 0, 0, 1, 1, 1, 1]])
        res = self.score(results)
        self.assertEqual(res[0], 1.)
        self.assertAlmostEqual(res[1], 5 / 6)
        res_target = self.score(results, target=1)
        self.assertEqual(res[0], res_target[0])
        self.assertEqual(res[1], res_target[1])
        res_target = self.score(results, target=0)
        self.assertEqual(res_target[0], 1.)
        self.assertAlmostEqual(res_target[1], 3 / 4)


class TestCA(unittest.TestCase):
    def test_init(self):
        res = Results(nmethods=2, nrows=100)
        res.actual[:50] = 0
        res.actual[50:] = 1
        res.predicted = np.vstack((res.actual, res.actual))
        np.testing.assert_almost_equal(CA(res), [1, 1])

        res.predicted[0][0] = 1
        np.testing.assert_almost_equal(CA(res), [0.99, 1])

        res.predicted[1] = 1 - res.predicted[1]
        np.testing.assert_almost_equal(CA(res), [0.99, 0])

    def test_call(self):
        res = Results(nmethods=2, nrows=100)
        res.actual[:50] = 0
        res.actual[50:] = 1
        res.predicted = np.vstack((res.actual, res.actual))
        ca = CA()
        np.testing.assert_almost_equal(ca(res), [1, 1])

        res.predicted[0][0] = 1
        np.testing.assert_almost_equal(ca(res), [0.99, 1])

        res.predicted[1] = 1 - res.predicted[1]
        np.testing.assert_almost_equal(ca(res), [0.99, 0])

    def test_bayes(self):
        x = np.random.randint(2, size=(100, 5))
        col = np.random.randint(5)
        y = x[:, col].reshape(100, 1).copy()
        t = Table.from_numpy(None, x, y)
        t = Discretize(
            method=discretize.EqualWidth(n=3))(t)
        nb = NaiveBayesLearner()
        res = TestOnTrainingData()(t, [nb])
        np.testing.assert_almost_equal(CA(res), [1])

        t = Table.from_numpy(None, t.X, t.Y.copy())
        with t.unlocked():
            t.Y[-20:] = 1 - t.Y[-20:]
        res = TestOnTrainingData()(t, [nb])
        self.assertGreaterEqual(CA(res)[0], 0.75)
        self.assertLess(CA(res)[0], 1)


class TestAUC(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.iris = Table('iris')

    def test_tree(self):
        tree = SklTreeLearner()
        res = CrossValidation(k=2)(self.iris, [tree])
        self.assertGreater(AUC(res)[0], 0.8)
        self.assertLess(AUC(res)[0], 1.)

    def test_constant_prob(self):
        maj = MajorityLearner()
        res = TestOnTrainingData()(self.iris, [maj])
        self.assertEqual(AUC(res)[0], 0.5)

    def test_multiclass_auc_multi_learners(self):
        learners = [LogisticRegressionLearner(),
                    MajorityLearner()]
        res = CrossValidation(k=10)(self.iris, learners)
        self.assertGreater(AUC(res)[0], 0.6)
        self.assertLess(AUC(res)[1], 0.6)
        self.assertGreater(AUC(res)[1], 0.4)

    def test_auc_on_multiclass_data_returns_1d_array(self):
        titanic = Table('titanic')[:100]
        lenses = Table(test_filename('datasets/lenses.tab'))[:100]
        majority = MajorityLearner()

        results = TestOnTrainingData()(lenses, [majority])
        auc = AUC(results)
        self.assertEqual(auc.ndim, 1)

        results = TestOnTrainingData()(titanic, [majority])
        auc = AUC(results)
        self.assertEqual(auc.ndim, 1)

    def test_auc_scores(self):
        actual = np.array([0., 0., 0., 1., 1., 1.])
        for predicted, auc in (([1., 1., 1., 0., 0., 0.], 0.),      # All wrong
                               ([0., 0., 0., 0., 0., 0.], 0.5),     # All with same probability
                               ([0., 0., 0., 1., 1., 1.], 1.),      # All correct
                               ([0., 0., 0., 1., 1., 0.], 5 / 6),   # One wrong
                               ([1., 1., 0., 1., 1., 1.], 4 / 6),   # Two wrong
                               ([1., 1., 0., 1., 1., 0.], 3 / 6)):  # Three wrong
            self.assertAlmostEqual(self.compute_auc(actual, predicted), auc)

    def compute_auc(self, actual, predicted):
        predicted = np.array(predicted).reshape(1, -1)
        probabilities = np.zeros((1, predicted.shape[1], 2))
        probabilities[0, :, 1] = predicted[0]
        probabilities[0, :, 0] = 1 - predicted[0]
        results = Results(
            nmethods=1, domain=Domain([], [DiscreteVariable("x", values='01')]),
            actual=actual, predicted=predicted)
        results.probabilities = probabilities
        return AUC(results)[0]


class TestLogLoss(unittest.TestCase):
    def test_log_loss(self):
        data = Table('iris')
        majority = MajorityLearner()
        results = TestOnTrainingData()(data, [majority])
        ll = LogLoss(results)
        self.assertAlmostEqual(ll[0], - np.log(1 / 3))

    def _log_loss(self, act, prob):
        ll = np.dot(np.log(prob[:, 0]), act[:, 0]) + \
             np.dot(np.log(prob[:, 1]), act[:, 1])
        return - ll / len(act)

    def test_log_loss_calc(self):
        data = Table('titanic')
        learner = LogisticRegressionLearner()
        results = TestOnTrainingData()(data, [learner])

        actual = np.copy(results.actual)
        actual = actual.reshape(actual.shape[0], 1)
        actual = np.hstack((1 - actual, actual))
        probab = results.probabilities[0]

        ll_calc = self._log_loss(actual, probab)
        ll_orange = LogLoss(results)
        self.assertAlmostEqual(ll_calc, ll_orange[0])


class TestSpecificity(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.iris = Table('iris')
        cls.score = Specificity()

    def test_specificity_iris(self):
        learner = LogisticRegressionLearner(preprocessors=[])
        res = TestOnTrainingData()(self.iris, [learner])
        self.assertGreaterEqual(
            self.score(res, average='weighted')[0], (1 + 0.99 + 0.95) / 3
        )
        self.assertGreaterEqual(
            self.score(res, target=1)[0], 99 / (99 + 1)
        )
        self.assertGreaterEqual(
            self.score(res, target=1, average=None)[0],  99 / (99 + 1)
        )
        self.assertGreaterEqual(
            self.score(res, target=1, average='weighted')[0], 99 / (99 + 1)
        )
        self.assertGreaterEqual(
            self.score(res, target=0, average=None)[0], 1
        )
        self.assertGreaterEqual(
            self.score(res, target=2, average=None)[0], 95 / (95 + 5)
        )

    def test_precision_multiclass(self):
        results = Results(
            domain=Domain([], DiscreteVariable(name="y", values="01234")),
            actual=[0, 4, 4, 1, 2, 0, 1, 2, 3, 2])
        results.predicted = np.array([[0, 4, 4, 1, 2, 0, 1, 2, 3, 2],
                                      [0, 1, 4, 1, 1, 0, 0, 2, 3, 1]])
        res = self.score(results, average='weighted')
        self.assertEqual(res[0], 1.)
        self.assertAlmostEqual(res[1], 0.9, 5)

        for target, prob in ((0, 7 / 8),
                             (1, 5 / 8),
                             (2, 1),
                             (3, 1),
                             (4, 1)):
            res = self.score(results, target=target, average=None)
            self.assertEqual(res[0], 1.)
            self.assertEqual(res[1], prob)

    def test_precision_binary(self):
        results = Results(
            domain=Domain([], DiscreteVariable(name="y", values="01")),
            actual=[0, 1, 1, 1, 0, 0, 1, 0, 0, 1])
        results.predicted = np.array([[0, 1, 1, 1, 0, 0, 1, 0, 0, 1],
                                      [0, 1, 1, 1, 0, 0, 1, 1, 1, 0]])
        res = self.score(results)
        self.assertEqual(res[0], 1.)
        self.assertAlmostEqual(res[1], 3 / 5)
        res_target = self.score(results, target=1)
        self.assertEqual(res[0], res_target[0])
        self.assertEqual(res[1], res_target[1])
        res_target = self.score(results, target=0)
        self.assertEqual(res_target[0], 1.)
        self.assertAlmostEqual(res_target[1], 4 / 5)

    def test_errors(self):
        learner = LogisticRegressionLearner(preprocessors=[])
        res = TestOnTrainingData()(self.iris, [learner])

        # binary average does not work for number of classes different than 2
        self.assertRaises(ValueError, self.score, res, average="binary")

        # implemented only weighted and binary averaging
        self.assertRaises(ValueError, self.score, res, average="abc")


if __name__ == '__main__':
    unittest.main()
    del TestScoreMetaType



"""
Adapted from a code editor component created
for Enki editor as replacement for QScintilla.
Copyright (C) 2020  Andrei Kopats

Originally licensed under the terms of GNU Lesser General Public License
as published by the Free Software Foundation, version 2.1 of the license.
This is compatible with Orange3's GPL-3.0 license.
"""
import time

from AnyQt.QtCore import Qt
from AnyQt.QtGui import QTextCursor, QColor
from AnyQt.QtWidgets import QTextEdit, QApplication

# Bracket highlighter.
# Calculates list of QTextEdit.ExtraSelection


class _TimeoutException(UserWarning):
    """Operation timeout happened
    """


class BracketHighlighter:
    """Bracket highliter.
    Calculates list of QTextEdit.ExtraSelection

    Currently, this class might be just a set of functions.
    Probably, it will contain instance specific selection colors later
    """
    MATCHED_COLOR = QColor('#0b0')
    UNMATCHED_COLOR = QColor('#a22')

    _MAX_SEARCH_TIME_SEC = 0.02

    _START_BRACKETS = '({['
    _END_BRACKETS = ')}]'
    _ALL_BRACKETS = _START_BRACKETS + _END_BRACKETS
    _OPOSITE_BRACKET = dict(zip(_START_BRACKETS + _END_BRACKETS, _END_BRACKETS + _START_BRACKETS))

    # instance variable. None or ((block, columnIndex), (block, columnIndex))
    currentMatchedBrackets = None

    def _iterateDocumentCharsForward(self, block, startColumnIndex):
        """Traverse document forward. Yield (block, columnIndex, char)
        Raise _TimeoutException if time is over
        """
        # Chars in the start line
        endTime = time.time() + self._MAX_SEARCH_TIME_SEC
        for columnIndex, char in list(enumerate(block.text()))[startColumnIndex:]:
            yield block, columnIndex, char
        block = block.next()

        # Next lines
        while block.isValid():
            for columnIndex, char in enumerate(block.text()):
                yield block, columnIndex, char

            if time.time() > endTime:
                raise _TimeoutException('Time is over')

            block = block.next()

    def _iterateDocumentCharsBackward(self, block, startColumnIndex):
        """Traverse document forward. Yield (block, columnIndex, char)
        Raise _TimeoutException if time is over
        """
        # Chars in the start line
        endTime = time.time() + self._MAX_SEARCH_TIME_SEC
        for columnIndex, char in reversed(list(enumerate(block.text()[:startColumnIndex]))):
            yield block, columnIndex, char
        block = block.previous()

        # Next lines
        while block.isValid():
            for columnIndex, char in reversed(list(enumerate(block.text()))):
                yield block, columnIndex, char

            if time.time() > endTime:
                raise _TimeoutException('Time is over')

            block = block.previous()

    def _findMatchingBracket(self, bracket, qpart, block, columnIndex):
        """Find matching bracket for the bracket.
        Return (block, columnIndex) or (None, None)
        Raise _TimeoutException, if time is over
        """
        if bracket in self._START_BRACKETS:
            charsGenerator = self._iterateDocumentCharsForward(block, columnIndex + 1)
        else:
            charsGenerator = self._iterateDocumentCharsBackward(block, columnIndex)

        depth = 1
        oposite = self._OPOSITE_BRACKET[bracket]
        for b, c_index, char in charsGenerator:
            if qpart.isCode(b, c_index):
                if char == oposite:
                    depth -= 1
                    if depth == 0:
                        return b, c_index
                elif char == bracket:
                    depth += 1
        return None, None

    def _makeMatchSelection(self, block, columnIndex, matched):
        """Make matched or unmatched QTextEdit.ExtraSelection
        """
        selection = QTextEdit.ExtraSelection()
        darkMode = QApplication.instance().property('darkMode')

        if matched:
            fgColor = self.MATCHED_COLOR
        else:
            fgColor = self.UNMATCHED_COLOR

        selection.format.setForeground(fgColor)
        # repaint hack
        selection.format.setBackground(Qt.white if not darkMode else QColor('#111111'))
        selection.cursor = QTextCursor(block)
        selection.cursor.setPosition(block.position() + columnIndex)
        selection.cursor.movePosition(QTextCursor.Right, QTextCursor.KeepAnchor)

        return selection

    def _highlightBracket(self, bracket, qpart, block, columnIndex):
        """Highlight bracket and matching bracket
        Return tuple of QTextEdit.ExtraSelection's
        """
        try:
            matchedBlock, matchedColumnIndex = self._findMatchingBracket(bracket, qpart,
                                                                         block, columnIndex)
        except _TimeoutException:  # not found, time is over
            return[] # highlight nothing

        if matchedBlock is not None:
            self.currentMatchedBrackets = ((block, columnIndex), (matchedBlock, matchedColumnIndex))
            return [self._makeMatchSelection(block, columnIndex, True),
                    self._makeMatchSelection(matchedBlock, matchedColumnIndex, True)]
        else:
            self.currentMatchedBrackets = None
            return [self._makeMatchSelection(block, columnIndex, False)]

    def extraSelections(self, qpart, block, columnIndex):
        """List of QTextEdit.ExtraSelection's, which highlighte brackets
        """
        blockText = block.text()

        if columnIndex < len(blockText) and \
             blockText[columnIndex] in self._ALL_BRACKETS and \
             qpart.isCode(block, columnIndex):
            return self._highlightBracket(blockText[columnIndex], qpart, block, columnIndex)
        elif columnIndex > 0 and \
           blockText[columnIndex - 1] in self._ALL_BRACKETS and \
           qpart.isCode(block, columnIndex - 1):
            return self._highlightBracket(blockText[columnIndex - 1], qpart, block, columnIndex - 1)
        else:
            self.currentMatchedBrackets = None
            return []



# Test methods with long descriptive names can omit docstrings
# pylint: disable=missing-docstring

import unittest
import warnings

import numpy as np
from scipy.sparse import csc_matrix, csr_matrix

import Orange
from Orange.clustering.kmeans import KMeans, KMeansModel
from Orange.data import Table, Domain, ContinuousVariable
from Orange.data.table import DomainTransformationError


class TestKMeans(unittest.TestCase):
    def setUp(self):
        self.kmeans = KMeans(n_clusters=2)
        self.iris = Orange.data.Table('iris')

    def test_kmeans(self):
        c = self.kmeans(self.iris)
        # First 20 iris belong to one cluster
        self.assertEqual(np.ndarray, type(c))
        self.assertEqual(len(self.iris), len(c))
        self.assertEqual(1, len(set(c[:20].ravel())))

    def test_kmeans_parameters(self):
        kmeans = KMeans(n_clusters=10, max_iter=10, random_state=42, tol=0.001,
                        init='random')
        c = kmeans(self.iris)
        self.assertEqual(np.ndarray, type(c))
        self.assertEqual(len(self.iris), len(c))

    def test_predict_table(self):
        c = self.kmeans(self.iris)
        self.assertEqual(np.ndarray, type(c))
        self.assertEqual(len(self.iris), len(c))

    def test_predict_numpy(self):
        c = self.kmeans.fit(self.iris.X)
        self.assertEqual(KMeansModel, type(c))
        self.assertEqual(np.ndarray, type(c.labels))
        self.assertEqual(len(self.iris), len(c.labels))

    def test_predict_sparse_csc(self):
        with self.iris.unlocked():
            self.iris.X = csc_matrix(self.iris.X[::20])
        c = self.kmeans(self.iris)
        self.assertEqual(np.ndarray, type(c))
        self.assertEqual(len(self.iris), len(c))

    def test_predict_spares_csr(self):
        with self.iris.unlocked():
            self.iris.X = csr_matrix(self.iris.X[::20])
        c = self.kmeans(self.iris)
        self.assertEqual(np.ndarray, type(c))
        self.assertEqual(len(self.iris), len(c))

    def test_model(self):
        c = self.kmeans.get_model(self.iris)
        self.assertEqual(KMeansModel, type(c))
        self.assertEqual(len(self.iris), len(c.labels))

        c1 = c(self.iris)
        # prediction of the model must be same since data are same
        np.testing.assert_array_almost_equal(c.labels, c1)

    def test_model_np(self):
        """
        Test with numpy array as an input in model.
        """
        c = self.kmeans.get_model(self.iris)
        c1 = c(self.iris.X)
        # prediction of the model must be same since data are same
        np.testing.assert_array_almost_equal(c.labels, c1)

    def test_model_sparse_csc(self):
        """
        Test with sparse array as an input in model.
        """
        c = self.kmeans.get_model(self.iris)
        c1 = c(csc_matrix(self.iris.X))
        # prediction of the model must be same since data are same
        np.testing.assert_array_almost_equal(c.labels, c1)

    def test_model_sparse_csr(self):
        """
        Test with sparse array as an input in model.
        """
        c = self.kmeans.get_model(self.iris)
        c1 = c(csr_matrix(self.iris.X))
        # prediction of the model must be same since data are same
        np.testing.assert_array_almost_equal(c.labels, c1)

    def test_model_instance(self):
        """
        Test with instance as an input in model.
        """
        c = self.kmeans.get_model(self.iris)
        c1 = c(self.iris[0])
        # prediction of the model must be same since data are same
        self.assertEqual(c1, c.labels[0])

    def test_model_list(self):
        """
        Test with list as an input in model.
        """
        c = self.kmeans.get_model(self.iris)
        c1 = c(self.iris.X.tolist())
        # prediction of the model must be same since data are same
        np.testing.assert_array_almost_equal(c.labels, c1)

        # example with a list of only one data item
        c1 = c(self.iris.X.tolist()[0])
        # prediction of the model must be same since data are same
        np.testing.assert_array_almost_equal(c.labels[0], c1)

    def test_model_bad_datatype(self):
        """
        Check model with data-type that is not supported.
        """
        c = self.kmeans.get_model(self.iris)
        self.assertRaises(TypeError, c, 10)

    def test_model_data_table_domain(self):
        """
        Check model with data-type that is not supported.
        """
        # ok domain
        data = Table(Domain(
            list(self.iris.domain.attributes) + [ContinuousVariable("a")]),
                     np.concatenate((self.iris.X, np.ones((len(self.iris), 1))), axis=1))
        c = self.kmeans.get_model(self.iris)
        res = c(data)
        np.testing.assert_array_almost_equal(c.labels, res)

        # totally different domain - should fail
        self.assertRaises(DomainTransformationError, c, Table("housing"))

    def test_deprecated_silhouette(self):
        with warnings.catch_warnings(record=True) as w:
            KMeans(compute_silhouette_score=True)

            assert len(w) == 1
            assert issubclass(w[-1].category, DeprecationWarning)

        with warnings.catch_warnings(record=True) as w:
            KMeans(compute_silhouette_score=False)

            assert len(w) == 1
            assert issubclass(w[-1].category, DeprecationWarning)



from Orange.base import Learner, Model, SklLearner, SklModel

__all__ = ["LearnerClassification", "ModelClassification",
           "SklModelClassification", "SklLearnerClassification"]


class LearnerClassification(Learner):

    def incompatibility_reason(self, domain):
        reason = None
        if len(domain.class_vars) > 1 and not self.supports_multiclass:
            reason = "Too many target variables."
        elif not domain.has_discrete_class:
            reason = "Categorical class variable expected."
        return reason


class ModelClassification(Model):
    def predict_proba(self, data):
        return self(data, ret=Model.Probs)


class SklModelClassification(SklModel, ModelClassification):
    pass


class SklLearnerClassification(SklLearner, LearnerClassification):
    __returns__ = SklModelClassification



# Test methods with long descriptive names can omit docstrings
# pylint: disable=missing-docstring

import unittest
import warnings

import numpy as np
from sklearn.exceptions import ConvergenceWarning

from Orange.data import Table
from Orange.classification import SGDClassificationLearner
from Orange.regression import SGDRegressionLearner
from Orange.evaluation import CrossValidation, RMSE, AUC


class TestSGDRegressionLearner(unittest.TestCase):
    def setUp(self):
        # Convergence warnings are irrelevant for these tests
        warnings.filterwarnings("ignore", ".*", ConvergenceWarning)
        super().setUp()

    def test_SGDRegression(self):
        nrows, ncols = 500, 5
        X = np.random.rand(nrows, ncols)
        y = X.dot(np.random.rand(ncols))
        data = Table.from_numpy(None, X, y)
        sgd = SGDRegressionLearner()
        cv = CrossValidation(k=3)
        res = cv(data, [sgd])
        self.assertLess(RMSE(res)[0], 0.1)

    def test_coefficients(self):
        lrn = SGDRegressionLearner()
        mod = lrn(Table("housing"))
        self.assertEqual(len(mod.coefficients), len(mod.domain.attributes))


class TestSGDClassificationLearner(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.iris = Table('iris')

    def setUp(self):
        # Convergence warnings are irrelevant for these tests
        warnings.filterwarnings("ignore", ".*", ConvergenceWarning)
        super().setUp()

    def test_SGDClassification(self):
        sgd = SGDClassificationLearner()
        cv = CrossValidation(k=3)
        res = cv(self.iris, [sgd])
        self.assertGreater(AUC(res)[0], 0.8)

    def test_coefficients(self):
        lrn = SGDClassificationLearner()
        mod = lrn(self.iris)
        self.assertEqual(len(mod.coefficients[0]), len(mod.domain.attributes))

    def test_predictions_shapes(self):
        """
        Test the resulting shapes of probabilities for SGD
        """
        # one where probabilities computed from data
        # hinge loss do not enable predict_proba
        lrn = SGDClassificationLearner()
        mod = lrn(self.iris)
        self.assertTupleEqual((50, 3), mod(self.iris[:50], mod.Probs).shape)
        self.assertTupleEqual((50,), mod(self.iris[:50], mod.Value).shape)

        # in this case probabilities are retrieved by skl_learner.predict_proba
        lrn = SGDClassificationLearner(loss='modified_huber')
        mod = lrn(self.iris)
        self.assertTupleEqual((50, 3), mod(self.iris[:50], mod.Probs).shape)
        self.assertTupleEqual((50,), mod(self.iris[:50], mod.Value).shape)



import unittest

import numpy as np
import pandas as pd

from Orange.data import (
    DiscreteVariable,
    ContinuousVariable,
    Domain,
    StringVariable,
    Table,
    table_to_frame,
)


def create_sample_data():
    domain = Domain(
        [
            ContinuousVariable("a"),
            ContinuousVariable("b"),
            ContinuousVariable("cvar"),
            DiscreteVariable("dvar", values=["val1", "val2"]),
        ],
        metas=[StringVariable("svar")],
    )
    return Table.from_numpy(
        domain,
        np.array(
            [
                [1, 1, 0.1, 0],
                [1, 1, 0.2, 1],
                [1, 2, np.nan, np.nan],
                [1, 2, 0.3, 1],
                [1, 3, 0.3, 0],
                [1, 3, 0.4, 1],
                [1, 3, 0.6, 0],
                [2, 1, 1.0, 1],
                [2, 1, 2.0, 0],
                [2, 2, 3.0, 1],
                [2, 2, -4.0, 0],
                [2, 3, 5.0, 1],
                [2, 3, 5.0, 0],
            ]
        ),
        metas=np.array(
            [
                ["sval1"],
                ["sval2"],
                [""],
                ["sval2"],
                ["sval1"],
                ["sval2"],
                ["sval1"],
                ["sval2"],
                ["sval1"],
                ["sval2"],
                ["sval1"],
                ["sval2"],
                ["sval1"],
            ]
        ),
    )


# pylint: disable=abstract-method
class AlternativeTable(Table):
    pass


class DomainTest(unittest.TestCase):
    def setUp(self) -> None:
        self.data = create_sample_data()

    def test_simple_aggregation(self):
        """Test aggregation results"""
        d = self.data.domain
        gb = self.data.groupby([d["a"]])
        output = gb.aggregate({d["a"]: ["mean"], d["b"]: ["mean"]})

        np.testing.assert_array_almost_equal(output.X, [[1, 2.143], [2, 2]], decimal=3)
        np.testing.assert_array_almost_equal(output.metas, [[1], [2]], decimal=3)
        self.assertListEqual(
            ["a - mean", "b - mean"], [d.name for d in output.domain.attributes]
        )
        self.assertListEqual(["a"], [d.name for d in output.domain.metas])

    def test_aggregation(self):
        d = self.data.domain
        gb = self.data.groupby([self.data.domain["a"], self.data.domain["b"]])
        output = gb.aggregate(
            {
                d["cvar"]: [("Mean", "mean"), ("Median", "median"), ("Mean1", np.mean)],
                d["dvar"]: [("Count defined", "count"), ("Count", "size")],
                d["svar"]: [("Concatenate", "".join)],
            }
        )

        expected_columns = [
            "cvar - Mean",
            "cvar - Median",
            "cvar - Mean1",
            "dvar - Count defined",
            "dvar - Count",
            "svar - Concatenate",
            "a",  # groupby variables are last two in metas
            "b",
        ]

        exp_df = pd.DataFrame(
            [
                [0.15, 0.15, 0.15, 2, 2, "sval1sval2", 1, 1],
                [0.3, 0.3, 0.3, 1, 2, "sval2", 1, 2],
                [0.433, 0.4, 0.433, 3, 3, "sval1sval2sval1", 1, 3],
                [1.5, 1.5, 1.5, 2, 2, "sval2sval1", 2, 1],
                [-0.5, -0.5, -0.5, 2, 2, "sval2sval1", 2, 2],
                [5, 5, 5, 2, 2, "sval2sval1", 2, 3],
            ],
            columns=expected_columns,
        )

        out_df = table_to_frame(output, include_metas=True)

        pd.testing.assert_frame_equal(
            out_df,
            exp_df,
            check_dtype=False,
            check_column_type=False,
            check_categorical=False,
            atol=1e-3,
        )

    def test_preserve_table_class(self):
        """
        Test whether result table has the same type than the imnput table,
        e.g. if input table corpus the resutlitn table must be corpus too.
        """
        data = AlternativeTable.from_table(self.data.domain, self.data)
        gb = data.groupby([data.domain["a"]])
        output = gb.aggregate({data.domain["a"]: ["mean"]})
        self.assertIsInstance(output, AlternativeTable)


if __name__ == "__main__":
    unittest.main()



# Pull members from modules to Orange.data namespace
# pylint: disable=wildcard-import

from .variable import *
from .instance import *
from .domain import *
from .storage import *
from .table import *
from .io_util import *
from .io import *
from .filter import *
from .pandas_compat import *
from .aggregate import *



# Test methods with long descriptive names can omit docstrings
# pylint: disable=missing-docstring

import unittest
import inspect
import pkgutil
import traceback

import Orange
from Orange.data import Table
from Orange.regression import Learner, CurveFitLearner
from Orange.tests import test_filename


def all_learners():
    regression_modules = pkgutil.walk_packages(
        path=Orange.regression.__path__,
        prefix="Orange.regression.",
        onerror=lambda x: None)
    for _, modname, _ in regression_modules:
        try:
            module = pkgutil.importlib.import_module(modname)
        except ImportError:
            continue

        for _, class_ in inspect.getmembers(module, inspect.isclass):
            if issubclass(class_, Learner) and 'base' not in class_.__module__:
                yield class_


def init_learner(learner, table):
    if learner == CurveFitLearner:
        return CurveFitLearner(
            lambda x, a: x[:, -1] * a, [],
            [table.domain.attributes[-1].name]
        )
    return learner()


class TestRegression(unittest.TestCase):
    def test_adequacy_all_learners(self):
        table = Table("iris")
        for learner in all_learners():
            learner = init_learner(learner, table)
            with self.assertRaises(ValueError):
                learner(table)

    def test_adequacy_all_learners_multiclass(self):
        table = Table(test_filename("datasets/test8.tab"))
        for learner in all_learners():
            learner = init_learner(learner, table)
            with self.assertRaises(ValueError):
                learner(table)

    def test_missing_class(self):
        table = Table(test_filename("datasets/imports-85.tab"))
        for learner in all_learners():
            learner = init_learner(learner, table)
            model = learner(table)
            model(table)


if __name__ == "__main__":
    unittest.main()



from collections import OrderedDict

from AnyQt.QtCore import Qt
from AnyQt.QtWidgets import QHBoxLayout, QGridLayout, QLabel, QWidget

from Orange.widgets.report import bool_str
from Orange.data import ContinuousVariable, StringVariable, Domain, Table
from Orange.modelling.linear import SGDLearner
from Orange.widgets import gui
from Orange.widgets.model.owlogisticregression import create_coef_table
from Orange.widgets.settings import Setting
from Orange.widgets.utils.owlearnerwidget import OWBaseLearner
from Orange.widgets.utils.signals import Output
from Orange.widgets.utils.widgetpreview import WidgetPreview

MAXINT = 2 ** 31 - 1


class OWSGD(OWBaseLearner):
    name = 'Stochastic Gradient Descent'
    description = 'Minimize an objective function using a stochastic ' \
                  'approximation of gradient descent.'
    icon = "icons/SGD.svg"
    replaces = [
        "Orange.widgets.regression.owsgdregression.OWSGDRegression",
    ]
    priority = 90
    keywords = ["sgd"]

    settings_version = 2

    LEARNER = SGDLearner

    class Outputs(OWBaseLearner.Outputs):
        coefficients = Output("Coefficients", Table, explicit=True)

    reg_losses = (
        ('Squared Loss', 'squared_error'),
        ('Huber', 'huber'),
        ('ε insensitive', 'epsilon_insensitive'),
        ('Squared ε insensitive', 'squared_epsilon_insensitive'))

    cls_losses = (
        ('Hinge', 'hinge'),
        ('Logistic regression', 'log'),
        ('Modified Huber', 'modified_huber'),
        ('Squared Hinge', 'squared_hinge'),
        ('Perceptron', 'perceptron')) + reg_losses

    #: Regularization methods
    penalties = (
        ('None', 'none'),
        ('Lasso (L1)', 'l1'),
        ('Ridge (L2)', 'l2'),
        ('Elastic Net', 'elasticnet'))

    learning_rates = (
        ('Constant', 'constant'),
        ('Optimal', 'optimal'),
        ('Inverse scaling', 'invscaling'))

    learner_name = Setting('SGD')
    #: Loss function index for classification problems
    cls_loss_function_index = Setting(0)
    #: Epsilon loss function parameter for classification problems
    cls_epsilon = Setting(.1)
    #: Loss function index for regression problems
    reg_loss_function_index = Setting(0)
    #: Epsilon loss function parameter for regression problems
    reg_epsilon = Setting(.1)

    penalty_index = Setting(2)
    #: Regularization strength
    alpha = Setting(1e-5)
    #: Elastic Net mixing parameter
    l1_ratio = Setting(.15)

    shuffle = Setting(True)
    use_random_state = Setting(False)
    random_state = Setting(0)
    learning_rate_index = Setting(0)
    eta0 = Setting(.01)
    power_t = Setting(.25)
    max_iter = Setting(1000)
    tol = Setting(1e-3)
    tol_enabled = Setting(True)

    def add_main_layout(self):
        main_widget = QWidget()
        layout = QHBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        main_widget.setLayout(layout)
        self.controlArea.layout().addWidget(main_widget)

        left = gui.vBox(main_widget).layout()
        right = gui.vBox(main_widget).layout()
        self._add_algorithm_to_layout(left)
        self._add_regularization_to_layout(left)
        self._add_learning_params_to_layout(right)

    def _foc_frame_width(self):
        style = self.style()
        return style.pixelMetric(style.PM_FocusFrameHMargin) + \
               style.pixelMetric(style.PM_ComboBoxFrameWidth)

    def _add_algorithm_to_layout(self, layout):
        # this is part of init, pylint: disable=attribute-defined-outside-init
        grid = QGridLayout()
        box = gui.widgetBox(None, 'Loss functions', orientation=grid)
        layout.addWidget(box)
        # Classfication loss function
        self.cls_loss_function_combo = gui.comboBox(
            None, self, 'cls_loss_function_index', orientation=Qt.Horizontal,
            items=list(zip(*self.cls_losses))[0],
            callback=self._on_cls_loss_change)
        hbox = gui.hBox(None)
        hbox.layout().addSpacing(self._foc_frame_width())
        self.cls_epsilon_spin = gui.spin(
            hbox, self, 'cls_epsilon', 0, 1., 1e-2, spinType=float,
            label='ε: ', controlWidth=80, alignment=Qt.AlignRight,
            callback=self.settings_changed)
        hbox.layout().addStretch()
        grid.addWidget(QLabel("Classification: "), 0, 0)
        grid.addWidget(self.cls_loss_function_combo, 0, 1)
        grid.addWidget(hbox, 1, 1)

        # Regression loss function
        self.reg_loss_function_combo = gui.comboBox(
            None, self, 'reg_loss_function_index', orientation=Qt.Horizontal,
            items=list(zip(*self.reg_losses))[0],
            callback=self._on_reg_loss_change)
        hbox = gui.hBox(None)
        hbox.layout().addSpacing(self._foc_frame_width())
        self.reg_epsilon_spin = gui.spin(
            hbox, self, 'reg_epsilon', 0, 1., 1e-2, spinType=float,
            label='ε: ', controlWidth=80, alignment=Qt.AlignRight,
            callback=self.settings_changed)
        hbox.layout().addStretch()
        grid.addWidget(QLabel("Regression: "), 2, 0)
        grid.addWidget(self.reg_loss_function_combo, 2, 1)
        grid.addWidget(hbox, 3, 1)

        # Enable/disable appropriate controls
        self._on_cls_loss_change()
        self._on_reg_loss_change()

    def _add_regularization_to_layout(self, layout):
        # this is part of init, pylint: disable=attribute-defined-outside-init
        box = gui.widgetBox(None, 'Regularization')
        layout.addWidget(box)
        hlayout = gui.hBox(box)
        self.penalty_combo = gui.comboBox(
            hlayout, self, 'penalty_index',
            items=list(zip(*self.penalties))[0], orientation=Qt.Horizontal,
            callback=self._on_regularization_change)
        self.l1_ratio_box = gui.spin(
            hlayout, self, 'l1_ratio', 0, 1., 1e-2, spinType=float,
            label='Mixing: ', controlWidth=80,
            alignment=Qt.AlignRight, callback=self.settings_changed).box

        hbox = gui.indentedBox(
            box, sep=self._foc_frame_width(), orientation=Qt.Horizontal)
        self.alpha_spin = gui.spin(
            hbox, self, 'alpha', 0, 10., .1e-4, spinType=float, controlWidth=80,
            label='Strength (α): ', alignment=Qt.AlignRight,
            callback=self.settings_changed)
        hbox.layout().addStretch()

        # Enable/disable appropriate controls
        self._on_regularization_change()

    def _add_learning_params_to_layout(self, layout):
        # this is part of init, pylint: disable=attribute-defined-outside-init
        box = gui.widgetBox(None, 'Optimization')
        layout.addWidget(box)
        self.learning_rate_combo = gui.comboBox(
            box, self, 'learning_rate_index', label='Learning rate: ',
            items=list(zip(*self.learning_rates))[0],
            orientation=Qt.Horizontal, callback=self._on_learning_rate_change)
        self.eta0_spin = gui.spin(
            box, self, 'eta0', 1e-4, 1., 1e-4, spinType=float,
            label='Initial learning rate (η<sub>0</sub>): ',
            alignment=Qt.AlignRight, controlWidth=80,
            callback=self.settings_changed)
        self.power_t_spin = gui.spin(
            box, self, 'power_t', 0, 1., 1e-4, spinType=float,
            label='Inverse scaling exponent (t): ',
            alignment=Qt.AlignRight, controlWidth=80,
            callback=self.settings_changed)
        gui.separator(box, height=12)

        self.max_iter_spin = gui.spin(
            box, self, 'max_iter', 1, MAXINT - 1, label='Number of iterations: ',
            controlWidth=80, alignment=Qt.AlignRight,
            callback=self.settings_changed)

        self.tol_spin = gui.spin(
            box, self, 'tol', 0, 10., .1e-3, spinType=float, controlWidth=80,
            label='Tolerance (stopping criterion): ', checked='tol_enabled',
            alignment=Qt.AlignRight, callback=self.settings_changed)
        gui.separator(box, height=12)

        # Wrap shuffle_cbx inside another hbox to align it with the random_seed
        # spin box on OSX
        self.shuffle_cbx = gui.checkBox(
            gui.hBox(box), self, 'shuffle',
            'Shuffle data after each iteration',
            callback=self._on_shuffle_change)
        self.random_seed_spin = gui.spin(
            box, self, 'random_state', 0, MAXINT,
            label='Fixed seed for random shuffling: ', controlWidth=80,
            alignment=Qt.AlignRight, callback=self.settings_changed,
            checked='use_random_state', checkCallback=self.settings_changed)

        # Enable/disable appropriate controls
        self._on_learning_rate_change()
        self._on_shuffle_change()

    def _on_cls_loss_change(self):
        # Epsilon parameter for classification loss
        if self.cls_losses[self.cls_loss_function_index][1] in (
                'huber', 'epsilon_insensitive', 'squared_epsilon_insensitive'):
            self.cls_epsilon_spin.setEnabled(True)
        else:
            self.cls_epsilon_spin.setEnabled(False)

        self.settings_changed()

    def _on_reg_loss_change(self):
        # Epsilon parameter for regression loss
        if self.reg_losses[self.reg_loss_function_index][1] in (
                'huber', 'epsilon_insensitive', 'squared_epsilon_insensitive'):
            self.reg_epsilon_spin.setEnabled(True)
        else:
            self.reg_epsilon_spin.setEnabled(False)

        self.settings_changed()

    def _on_regularization_change(self):
        # Alpha parameter
        if self.penalties[self.penalty_index][1] in ('l1', 'l2', 'elasticnet'):
            self.alpha_spin.setEnabled(True)
        else:
            self.alpha_spin.setEnabled(False)
        # Elastic Net mixing parameter
        self.l1_ratio_box.setHidden(
            self.penalties[self.penalty_index][1] != 'elasticnet')

        self.settings_changed()

    def _on_learning_rate_change(self):
        # Eta_0 parameter
        if self.learning_rates[self.learning_rate_index][1] in \
                ('constant', 'invscaling'):
            self.eta0_spin.setEnabled(True)
        else:
            self.eta0_spin.setEnabled(False)
        # Power t parameter
        if self.learning_rates[self.learning_rate_index][1] in \
                ('invscaling',):
            self.power_t_spin.setEnabled(True)
        else:
            self.power_t_spin.setEnabled(False)

        self.settings_changed()

    def _on_shuffle_change(self):
        if self.shuffle:
            self.random_seed_spin[0].setEnabled(True)
        else:
            self.use_random_state = False
            self.random_seed_spin[0].setEnabled(False)

        self.settings_changed()

    def create_learner(self):
        params = {}
        if self.use_random_state:
            params['random_state'] = self.random_state

        return self.LEARNER(
            classification_loss=self.cls_losses[self.cls_loss_function_index][1],
            classification_epsilon=self.cls_epsilon,
            regression_loss=self.reg_losses[self.reg_loss_function_index][1],
            regression_epsilon=self.reg_epsilon,
            penalty=self.penalties[self.penalty_index][1],
            alpha=self.alpha,
            l1_ratio=self.l1_ratio,
            shuffle=self.shuffle,
            learning_rate=self.learning_rates[self.learning_rate_index][1],
            eta0=self.eta0,
            power_t=self.power_t,
            max_iter=self.max_iter,
            tol=self.tol if self.tol_enabled else None,
            preprocessors=self.preprocessors,
            **params)

    def get_learner_parameters(self):
        params = OrderedDict({})
        # Classification loss function
        params['Classification loss function'] = self.cls_losses[
            self.cls_loss_function_index][0]
        if self.cls_losses[self.cls_loss_function_index][1] in (
                'huber', 'epsilon_insensitive', 'squared_epsilon_insensitive'):
            params['Epsilon (ε) for classification'] = self.cls_epsilon
        # Regression loss function
        params['Regression loss function'] = self.reg_losses[
            self.reg_loss_function_index][0]
        if self.reg_losses[self.reg_loss_function_index][1] in (
                'huber', 'epsilon_insensitive', 'squared_epsilon_insensitive'):
            params['Epsilon (ε) for regression'] = self.reg_epsilon

        params['Regularization'] = self.penalties[self.penalty_index][0]
        # Regularization strength
        if self.penalties[self.penalty_index][1] in ('l1', 'l2', 'elasticnet'):
            params['Regularization strength (α)'] = self.alpha
        # Elastic Net mixing
        if self.penalties[self.penalty_index][1] in ('elasticnet',):
            params['Elastic Net mixing parameter (L1 ratio)'] = self.l1_ratio

        params['Learning rate'] = self.learning_rates[
            self.learning_rate_index][0]
        # Eta
        if self.learning_rates[self.learning_rate_index][1] in \
                ('constant', 'invscaling'):
            params['Initial learning rate (η<sub>0</sub>)'] = self.eta0
        # t
        if self.learning_rates[self.learning_rate_index][1] in \
                ('invscaling',):
            params['Inverse scaling exponent (t)'] = self.power_t

        params['Shuffle data after each iteration'] = bool_str(self.shuffle)
        if self.use_random_state:
            params['Random seed for shuffling'] = self.random_state

        return list(params.items())

    def update_model(self):
        super().update_model()
        coeffs = None
        if self.model is not None:
            if self.model.domain.class_var.is_discrete:
                coeffs = create_coef_table(self.model)
            else:
                attrs = [ContinuousVariable("coef")]
                domain = Domain(attrs, metas=[StringVariable("name")])
                cfs = list(self.model.intercept) + list(self.model.coefficients)
                names = ["intercept"] + \
                        [attr.name for attr in self.model.domain.attributes]
                coeffs = Table.from_list(domain, list(zip(cfs, names)))
                coeffs.name = "coefficients"
        self.Outputs.coefficients.send(coeffs)

    @classmethod
    def migrate_settings(cls, settings, version):
        if version < 2:
            settings["max_iter"] = settings.pop("n_iter", 5)
            settings["tol_enabled"] = False


if __name__ == "__main__":  # pragma: no cover
    WidgetPreview(OWSGD).run(Table("iris"))



# Test methods with long descriptive names can omit docstrings
# pylint: disable=missing-docstring,unsubscriptable-object,protected-access
import unittest
from functools import partial
from unittest.mock import patch, Mock

import numpy as np

from AnyQt.QtCore import QPoint, Qt, QModelIndex
from AnyQt.QtWidgets import QWidget, QApplication, QStyleOptionViewItem
from AnyQt.QtGui import QIcon

from orangewidget.settings import Context

from Orange.data import Table, ContinuousVariable, TimeVariable, Domain
from Orange.preprocess.discretize import TooManyIntervals
from Orange.widgets.data.owdiscretize import OWDiscretize, \
    IncreasingNumbersListValidator, VarHint, Methods, DefaultKey, \
    _fixed_width_discretization, _fixed_time_width_discretization, \
    _custom_discretization, variable_key, Options, DefaultHint, \
    _mdl_discretization, ListViewSearch, format_desc, DefaultDiscModel
from Orange.widgets.tests.base import WidgetTest, GuiTest
from Orange.widgets.utils.itemmodels import select_rows


class DataMixin:
    def prepare_data(self):
        self.domain = Domain([ContinuousVariable("x"),
                              ContinuousVariable("y"),
                              ContinuousVariable("z"),
                              TimeVariable("t"),
                              TimeVariable("u")])
        self.data = Table.from_numpy(self.domain, np.arange(20).reshape(4, 5))
        self.var_hints = {
            DefaultKey: VarHint(Methods.Keep, ()),
            ("x", False): VarHint(Methods.EqualFreq, (3, )),
            ("y", False): VarHint(Methods.Keep, ()),
            ("z", False): VarHint(Methods.Remove, ()),
            ("t", True): VarHint(Methods.Binning, (2, ))
        }
        # Copy the following line to tests, for reference:
        # Def: Keep, x: EqFreq 3, y: Keep, z: Remove, t (time): Bin 2, u (time):

class TestOWDiscretize(WidgetTest, DataMixin):
    def setUp(self):
        super().setUp()
        self.prepare_data()
        self.widget = self.create_widget(OWDiscretize)


    def test_empty_data(self):
        data = Table("iris")
        widget = self.widget
        self.send_signal(self.widget.Inputs.data,
                         Table.from_domain(data.domain))
        for m in range(len(Methods)):
            widget.var_hints = {DefaultKey: VarHint(m, ())}
            widget.commit.now()
            self.assertIsNotNone(self.get_output(widget.Outputs.data))

    def test_report(self):
        data = Table("brown-selected")

        w = self.create_widget(
            OWDiscretize,
            {"var_hints":
                {None: VarHint(Methods.EqualFreq, (3,)),
                 ('alpha 0', False): VarHint(Methods.Keep, ()),
                 ('alpha 7', False): VarHint(Methods.Remove, ()),
                 ('alpha 14', False): VarHint(Methods.Binning, (2, )),
                 ('alpha 21', False): VarHint(Methods.FixedWidth, ("0.05", )),
                 ('alpha 28', False): VarHint(Methods.EqualFreq, (4, )),
                 ('alpha 35', False): VarHint(Methods.MDL, ()),
                 ('alpha 42', False): VarHint(Methods.Custom, ("0, 0.125", )),
                 ('alpha 49', False): VarHint(Methods.MDL, ())},
             "__version__": 3})
        self.send_signal(w.Inputs.data, data)

        self.widget.send_report()

    def test_all(self):
        data = Table("brown-selected")

        w = self.create_widget(
            OWDiscretize,
            {"var_hints":
                {None: VarHint(Methods.EqualFreq, (3,)),
                 ('alpha 0', False): VarHint(Methods.Keep, ()),
                 ('alpha 7', False): VarHint(Methods.Remove, ()),
                 ('alpha 14', False): VarHint(Methods.Binning, (2, )),
                 ('alpha 21', False): VarHint(Methods.FixedWidth, ("0.05", )),
                 ('alpha 28', False): VarHint(Methods.EqualFreq, (4, )),
                 ('alpha 35', False): VarHint(Methods.MDL, ()),
                 ('alpha 42', False): VarHint(Methods.Custom, ("0, 0.125", )),
                 ('alpha 49', False): VarHint(Methods.MDL, ())},
             "__version__": 3})

        self.send_signal(w.Inputs.data, data)

        self.assertTrue(w.button_group.button(Methods.MDL).isEnabled())
        self.assertEqual(w.varview.default_view.model().hint,
                         VarHint(Methods.EqualFreq, (3, )))

        out = self.get_output(w.Outputs.data)
        dom = out.domain
        self.assertIsInstance(dom["alpha 0"], ContinuousVariable)
        self.assertNotIn("alpha 7", dom)
        self.assertEqual(dom["alpha 14"].values, ('< 0', '≥ 0'))
        self.assertEqual(dom["alpha 21"].values,
                         ('< -0.15', "-0.15 - -0.10", "-0.10 - -0.05",
                          "-0.05 - 0.00", "0.00 - 0.05", "0.05 - 0.10",
                          '≥ 0.10'))
        self.assertEqual(len(dom["alpha 28"].values), 4)
        self.assertNotIn("alpha 35", dom)  # removed by MDL
        self.assertEqual(dom["alpha 42"].values, ('< 0', '0 - 0.125', '≥ 0.125'))
        self.assertEqual(len(dom["alpha 49"].values), 2)

        self.send_signal(w.Inputs.data, None)
        self.assertIsNone(self.get_output(w.Outputs.data))
        self.assertIsNone(w.data)
        self.assertEqual(w.discretized_vars, {})
        self.assertEqual(len(w.varview.model()), 0)

        self.send_signal(w.Inputs.data, data)
        self.assertIsNotNone(self.get_output(w.Outputs.data))
        w.button_group.button(Methods.MDL).setChecked(True)
        self.assertTrue(w.button_group.button(Methods.MDL).isEnabled())
        self.assertTrue(w.button_group.button(Methods.MDL).isChecked())

        self.send_signal(w.Inputs.data, data[:, 0])
        self.assertFalse(w.button_group.button(Methods.MDL).isEnabled())
        self.assertFalse(w.button_group.button(Methods.MDL).isChecked())

        self.send_signal(w.Inputs.data, data)
        self.assertTrue(w.button_group.button(Methods.MDL).isEnabled())

    def test_get_values(self):
        w = self.widget

        w.binning_spin.setValue(5)
        w.width_line.setText("6")
        w.width_time_line.setText("7")
        w.width_time_unit.setCurrentIndex(1)
        w.freq_spin.setValue(8)
        w.width_spin.setValue(9)
        w.threshold_line.setText("1, 2, 3, 4, 5")

        self.assertEqual(w._get_values(Methods.Keep), ())
        self.assertEqual(w._get_values(Methods.Remove), ())
        self.assertEqual(w._get_values(Methods.Binning), (5, ))
        self.assertEqual(w._get_values(Methods.FixedWidth), ("6", ))
        self.assertEqual(w._get_values(Methods.FixedWidthTime), ("7", 1))
        self.assertEqual(w._get_values(Methods.EqualFreq), (8, ))
        self.assertEqual(w._get_values(Methods.EqualWidth), (9, ))
        self.assertEqual(w._get_values(Methods.MDL), ())
        self.assertEqual(w._get_values(Methods.Custom), ("1, 2, 3, 4, 5", ))

    def test_set_values(self):
        w = self.widget

        w._set_values(Methods.Keep, ())
        w._set_values(Methods.Remove, ())
        w._set_values(Methods.Binning, (5,))
        w._set_values(Methods.FixedWidth, ("6",))
        w._set_values(Methods.FixedWidthTime, ("7", 1))
        w._set_values(Methods.EqualFreq, (8,))
        w._set_values(Methods.EqualWidth, (9,))
        w._set_values(Methods.MDL, ())
        w._set_values(Methods.Custom, ("1, 2, 3, 4, 5",))

        self.assertEqual(w.binning_spin.value(), 5)
        self.assertEqual(w.width_line.text(), "6")
        self.assertEqual(w.width_time_line.text(), "7")
        self.assertEqual(w.width_time_unit.currentIndex(), 1)
        self.assertEqual(w.freq_spin.value(), 8)
        self.assertEqual(w.width_spin.value(), 9)
        self.assertEqual(w.threshold_line.text(), "1, 2, 3, 4, 5")

    def test_varkeys_for_selection(self):
        w = self.widget
        self.send_signal(w.Inputs.data, self.data)
        select_rows(w.varview, (0, 4))
        self.assertEqual(w.varkeys_for_selection(), [("x", False), ("u", True)])

    def test_change_selection_update_interface(self):
        w = self.widget
        self.send_signal(w.Inputs.data, self.data)
        w.var_hints = {
            DefaultKey: DefaultHint,
            ("x", False): VarHint(Methods.FixedWidth, ("10", )),
            ("y", False): VarHint(Methods.FixedWidth, ("10", )),
            ("z", False): VarHint(Methods.FixedWidth, ("5", )),
            ("t", False): VarHint(Methods.Binning, (5, ))
        }

        select_rows(w.varview, (0, 1))
        self.assertTrue(w.button_group.button(Methods.FixedWidth).isChecked())
        self.assertTrue(w.button_group.button(Methods.FixedWidth).isEnabled())
        self.assertFalse(w.button_group.button(Methods.FixedWidthTime).isEnabled())
        self.assertTrue(w.button_group.button(Methods.Custom).isEnabled())
        self.assertTrue(w.copy_to_custom.isEnabled())
        self.assertEqual(w.width_line.text(), "10")

        select_rows(w.varview, (1, 2))
        self.assertFalse(w.button_group.button(Methods.FixedWidth).isChecked())
        self.assertTrue(w.button_group.button(Methods.FixedWidth).isEnabled())
        self.assertFalse(w.button_group.button(Methods.FixedWidthTime).isEnabled())
        self.assertTrue(w.button_group.button(Methods.Custom).isEnabled())
        self.assertTrue(w.copy_to_custom.isEnabled())

        select_rows(w.varview, (2, 4))
        self.assertFalse(w.button_group.button(Methods.FixedWidth).isChecked())
        self.assertFalse(w.button_group.button(Methods.FixedWidth).isEnabled())
        self.assertFalse(w.button_group.button(Methods.FixedWidthTime).isEnabled())
        self.assertFalse(w.button_group.button(Methods.Custom).isEnabled())

        select_rows(w.varview, (3, 4))
        self.assertFalse(w.button_group.button(Methods.FixedWidth).isChecked())
        self.assertFalse(w.button_group.button(Methods.FixedWidth).isEnabled())
        self.assertTrue(w.button_group.button(Methods.FixedWidthTime).isEnabled())
        self.assertFalse(w.button_group.button(Methods.Custom).isEnabled())
        self.assertFalse(w.copy_to_custom.isEnabled())

        select_rows(w.varview.default_view, (0, ))
        self.assertEqual(len(w.varview.selectionModel().selectedIndexes()), 0)
        self.assertFalse(w.button_group.button(Methods.FixedWidth).isChecked())
        self.assertTrue(w.button_group.button(Methods.FixedWidth).isEnabled())
        self.assertTrue(w.button_group.button(Methods.FixedWidthTime).isEnabled())
        self.assertTrue(w.button_group.button(Methods.Custom).isEnabled())
        self.assertFalse(w.copy_to_custom.isEnabled())
        self.assertFalse(w.button_group.button(Methods.Default).isEnabled())
        w._check_button(Methods.FixedWidth, True)
        self.assertTrue(w.button_group.button(Methods.FixedWidth).isChecked())

        select_rows(w.varview, (3, ))
        self.assertEqual(len(w.varview.default_view.selectionModel().selectedIndexes()), 0)
        self.assertFalse(w.button_group.button(Methods.FixedWidth).isChecked())
        self.assertFalse(w.button_group.button(Methods.FixedWidth).isEnabled())
        self.assertTrue(w.button_group.button(Methods.FixedWidthTime).isEnabled())
        self.assertFalse(w.button_group.button(Methods.Custom).isEnabled())
        self.assertTrue(w.button_group.button(Methods.Default).isEnabled())

    def test_update_hints(self):
        w = self.widget
        update_disc = w._update_discretizations
        w._update_discretizations = Mock()
        w.width_line.setText("10")
        self.send_signal(w.Inputs.data, self.data)
        w.var_hints = {
            DefaultKey: DefaultHint,
            ("x", False): VarHint(Methods.EqualFreq, (3, )),
            ("y", False): VarHint(Methods.EqualFreq, (3, )),
            ("z", False): VarHint(Methods.EqualFreq, (4, )),
            ("t", True): VarHint(Methods.Binning, (5, ))
        }
        update_disc()
        self.assertEqual(len(w.discretized_vars), 5)

        select_rows(w.varview, (0, ))
        w.button_group.button(Methods.Default).click()
        self.assertNotIn(("x", False), w.var_hints)
        # Check that "x" is invalidated
        self.assertEqual(len(w.discretized_vars), 4)
        self.assertNotIn(("x", False), w.discretized_vars)
        update_disc()
        self.assertEqual(len(w.discretized_vars), 5)
        self.assertIn(("x", False), w.discretized_vars)

        select_rows(w.varview, (0, 1))
        w.button_group.button(Methods.FixedWidth).click()
        self.assertEqual(w.var_hints[("x", False)],
                         VarHint(Methods.FixedWidth, ("10", )))
        self.assertEqual(w.var_hints[("y", False)],
                         VarHint(Methods.FixedWidth, ("10", )))
        # Check that "x" and "y" are invalidated
        self.assertEqual(len(w.discretized_vars), 3)
        self.assertNotIn(("x", False), w.discretized_vars)
        self.assertNotIn(("y", False), w.discretized_vars)
        update_disc()
        self.assertEqual(len(w.discretized_vars), 5)
        self.assertIn(("x", False), w.discretized_vars)
        self.assertIn(("y", False), w.discretized_vars)

        w.width_line.setText("5")
        self.assertEqual(w.var_hints[("x", False)],
                         VarHint(Methods.FixedWidth, ("5", )))
        self.assertEqual(w.var_hints[("y", False)],
                         VarHint(Methods.FixedWidth, ("5", )))
        # Check that "x" and "y" are invalidated
        self.assertEqual(len(w.discretized_vars), 3)
        self.assertNotIn(("x", False), w.discretized_vars)
        self.assertNotIn(("y", False), w.discretized_vars)
        update_disc()
        self.assertEqual(len(w.discretized_vars), 5)
        self.assertIn(("x", False), w.discretized_vars)
        self.assertIn(("y", False), w.discretized_vars)

        select_rows(w.varview.default_view, (0, ))
        w.button_group.button(Methods.FixedWidth).click()
        self.assertEqual(len(w.discretized_vars), 4)
        self.assertNotIn(("u", True), w.discretized_vars)
        update_disc()
        self.assertEqual(len(w.discretized_vars), 5)
        self.assertIn(("u", True), w.discretized_vars)

    def test_discretize_var(self):
        w = self.widget
        self.send_signal(w.Inputs.data, self.data)

        x = self.data.domain["x"]
        t = self.data.domain["t"]

        s, dvar = w._discretize_var(x, VarHint(Methods.FixedWidthTime, ("10", 0)))
        self.assertIn("keep", s)
        self.assertIs(dvar, x)

        s, dvar = w._discretize_var(t, VarHint(Methods.FixedWidth, ("10", )))
        self.assertIn("keep", s)
        self.assertIs(dvar, t)

        try:
            Options[42] = Mock()

            # Errored
            # Unit test - mocked function
            Options[42].function = lambda *_: "foo error"
            s, dvar = w._discretize_var(t, VarHint(42, ()))
            self.assertIn("foo error", s)
            self.assertIsNone(dvar)
            # Real error
            s, dvar = w._discretize_var(t, VarHint(Methods.MDL, ()))
            self.assertIn("<", s)
            self.assertIsNone(dvar)

            # Removed attribute
            Options[42].function = lambda *_: None
            s, dvar = w._discretize_var(t, VarHint(42, ()))
            self.assertEqual("", s)
            self.assertIsNone(dvar)
            # Really removed
            s, dvar = w._discretize_var(t, VarHint(Methods.Remove, ()))
            self.assertEqual("", s)
            self.assertIsNone(dvar)

            # No intervals
            var = Mock(compute_value=Mock(points=[]))
            Options[42].function = lambda *_: var
            s, dvar = w._discretize_var(t, VarHint(42, ()))
            self.assertIn("removed", s)
            self.assertIsNone(dvar)
            s, dvar = w._discretize_var(x, VarHint(Methods.FixedWidth, ("1000", )))
            self.assertIn("removed", s)
            self.assertIsNone(dvar)

            # All fine
            var = Mock(compute_value=Mock(points=[1, 2, 3]))
            Options[42].function = lambda *_: var
            s, dvar = w._discretize_var(t, VarHint(42, ()))
            self.assertIn("1, 2, 3", s)
            self.assertIs(dvar, var)
            s, dvar = w._discretize_var(x, VarHint(Methods.EqualWidth, (3, )))
            self.assertEqual(dvar.compute_value.points, [5, 10])

        finally:
            del Options[42]

    def test_update_discretizations(self):
        w = self.widget
        # Def: Keep, x: EqFreq 3, y: Keep, z: Remove, t (time): Bin 2, u (time):
        w.var_hints = self.var_hints
        y, t, u = map(self.domain.__getitem__, "ytu")

        # no data: do nothing, but don't crash
        w._update_discretizations()

        self.send_signal(w.Inputs.data, self.data)
        d = w.discretized_vars
        self.assertEqual(len(d), 5)
        self.assertEqual(len(d[("x", False)].values), 3)
        self.assertIs(d[("y", False)], y)
        self.assertIsNone(d[("z", False)])
        self.assertIsNot(d[("t", True)], t)
        self.assertIsNotNone(d[("t", True)], t)
        self.assertIs(d[("u", True)], u)

        d[("t", True)] = t
        del d[("x", False)]
        del d[("u", True)]
        w._update_discretizations()
        self.assertEqual(len(d[("x", False)].values), 3)
        self.assertIs(d[("t", True)], t)
        self.assertIs(d[("u", True)], u)

        w.var_hints[None] = VarHint(Methods.Remove, ())
        del d[("u", True)]
        w._update_discretizations()
        self.assertIsNone(d[("u", True)])

    def test_copy_to_manual(self):
        w = self.widget
        w.var_hints = { DefaultKey: VarHint(Methods.EqualFreq, (5, )) }
        self.send_signal(w.Inputs.data, self.data)
        w.button_group.button(Methods.MDL).setChecked(True)

        select_rows(w.varview, (0, 2))
        self.assertTrue(w.copy_to_custom.isEnabled())
        w.copy_to_custom.click()
        self.assertFalse(any(w.button_group.button(i).isChecked()
                             for i in Methods))
        self.assertEqual(w.var_hints[("x", False)],
                         VarHint(Methods.Custom, ('2.5, 7.5, 12.5', )))
        self.assertEqual(w.var_hints[("z", False)],
                         VarHint(Methods.Custom, ('4.5, 9.5, 14.5', )))
        self.assertNotIn(("y", False), w.var_hints)

        select_rows(w.varview, (1, ))
        self.assertTrue(w.copy_to_custom.isEnabled())
        w.copy_to_custom.click()
        self.assertTrue(w.button_group.button(Methods.Custom).isChecked())
        self.assertEqual(w.var_hints[("x", False)],
                         VarHint(Methods.Custom, ('2.5, 7.5, 12.5', )))
        self.assertEqual(w.var_hints[("z", False)],
                         VarHint(Methods.Custom, ('4.5, 9.5, 14.5', )))
        self.assertEqual(w.var_hints[("y", False)],
                         VarHint(Methods.Custom, ('3.5, 8.5, 13.5', )))
        self.assertEqual(w.threshold_line.text(), '3.5, 8.5, 13.5')

        select_rows(w.varview, (1, 4))
        w.copy_to_custom.click()
        self.assertNotIn(("u", False), w.var_hints)

    def test_migration_2_3(self):
        # Obsolete, don't want to cause confusion by public import
        # pylint: disable=import-outside-toplevel
        from Orange.widgets.data.owdiscretize import \
            Default, EqualFreq, Leave, Custom, MDL, EqualWidth, DState
        context_values = {
            'saved_var_states':
                ({(2, 'age'): DState(method=Leave()),
                  (2, 'rest SBP'): DState(method=EqualWidth(k=4)),
                  (2, 'cholesterol'): DState(method=EqualFreq(k=6)),
                  (4, 'max HR'): DState(
                      method=Custom(points=(1.0, 2.0, 3.0))),
                  (2, 'ST by exercise'): DState(method=MDL()),
                  (2, 'major vessels colored'):
                      DState(method=Default(method=EqualFreq(k=3)))}, -2),
            '__version__': 2}

        settings = {'autosend': True, 'controlAreaVisible': True,
                    'default_cutpoints': (), 'default_k': 3,
                    'default_method_name': 'EqualFreq',
                    '__version__': 2,
                    "context_settings": [Context(values=context_values)]}

        OWDiscretize.migrate_settings(settings, 2)
        self.assertNotIn("default_method_name", settings)
        self.assertNotIn("default_k", settings)
        self.assertNotIn("default_cutpoints", settings)
        self.assertNotIn("context_settings", settings)
        self.assertEqual(
            settings["var_hints"],
            {None: VarHint(Methods.EqualFreq, (3,)),
             ('ST by exercise', False): VarHint(Methods.MDL, ()),
             ('age', False): VarHint(Methods.Keep, ()),
             ('cholesterol', False): VarHint(Methods.EqualFreq, (6,)),
             ('max HR', True): VarHint(Methods.Custom, (('1, 2, 3'),)),
             ('rest SBP', False): VarHint(Methods.EqualWidth, (4,))})


class TestValidator(unittest.TestCase):
    def test_validate(self):
        v = IncreasingNumbersListValidator()
        self.assertEqual(v.validate("", 0), (v.Intermediate, '', 0))
        self.assertEqual(v.validate("1", 1), (v.Acceptable, '1', 1))
        self.assertEqual(v.validate(",", 0), (v.Intermediate, ',', 0))
        self.assertEqual(v.validate("-", 0), (v.Intermediate, '-', 0))
        self.assertEqual(v.validate("1,,", 1), (v.Intermediate, '1,,', 1))
        self.assertEqual(v.validate("1,a,", 1), (v.Invalid, '1,a,', 3))
        self.assertEqual(v.validate("a", 1), (v.Invalid, 'a', 1))
        self.assertEqual(v.validate("1,1", 0), (v.Intermediate, '1,1', 0))
        self.assertEqual(v.validate("1,12", 0), (v.Acceptable, '1,12', 0))

        self.assertEqual(v.validate("1, 2 ", 5), (v.Intermediate, "1, 2, ", 6))


class TestModels(WidgetTest, DataMixin):
    def setUp(self):
        self.prepare_data()
        self.widget = self.create_widget(OWDiscretize)

    def test_delegate(self):
        self.prepare_data()
        w = self.widget
        w.var_hints = self.var_hints
        # Def: Keep, x: EqFreq 3, y: Keep, z: Remove, t (time): Bin 2, u (time):
        self.send_signal(w.Inputs.data, self.data)

        model = w.varview.model()
        delegate: ListViewSearch.DiscDelegate = w.varview.itemDelegate()
        option = QStyleOptionViewItem()
        delegate.initStyleOption(option, model.index(0))
        self.assertTrue(option.font.bold())

        option = QStyleOptionViewItem()
        delegate.initStyleOption(option, model.index(4))
        self.assertFalse(option.font.bold())

    def test_layout(self):
        # Not much to test, just don't crash
        self.widget.varview.updateGeometries()

    def test_model(self):
        self.prepare_data()
        w = self.widget
        w.var_hints = self.var_hints
        # Def: Keep, x: EqFreq 3, y: Keep, z: Remove, t (time): Bin 2, u (time):
        self.send_signal(w.Inputs.data, self.data)

        model = w.varview.model()
        display = model.index(0).data()
        self.assertIn("x", display)
        self.assertIn("equal", display)
        self.assertIn("3", display)
        self.assertIn(
            str(w.discretized_vars[("x", False)].compute_value.points[0])[:3],
            display)

        tooltip = model.index(0).data(Qt.ToolTipRole)
        self.assertIn("x", tooltip)
        self.assertIn(
            str(w.discretized_vars[("x", False)].compute_value.points[0])[:3],
            tooltip)

        display = model.index(1).data()
        self.assertIn("y", display)
        self.assertIn("keep", display)

        self.assertIsNone(model.index(1).data(Qt.ToolTipRole))

        w.var_hints[("x", False)] = VarHint(Methods.EqualWidth, (7, ))
        del w.discretized_vars[("x", False)]
        w._update_discretizations()
        display = model.index(0).data()
        self.assertIn("x", display)
        self.assertIn("equal", display)
        self.assertIn("3", display)
        self.assertIn(
            str(w.discretized_vars[("x", False)].compute_value.points[0])[:3],
            display)


class TestDefaultDiscModel(GuiTest):
    def test_counts(self):
        model = DefaultDiscModel()
        self.assertEqual(model.rowCount(QModelIndex()), 1)
        self.assertEqual(model.rowCount(model.index(0)), 0)

        self.assertEqual(model.columnCount(QModelIndex()), 1)
        self.assertEqual(model.columnCount(model.index(0)), 0)

    def test_data(self):
        model = DefaultDiscModel()
        self.assertIn(format_desc(DefaultHint), model.index(0).data())
        self.assertIsInstance(model.index(0).data(Qt.DecorationRole), QIcon)
        self.assertIsInstance(model.index(0).data(Qt.ToolTipRole), str)

        hint = VarHint(Methods.FixedWidth, ("314", ))
        model.setData(model.index(0), hint, Qt.UserRole)
        self.assertIn(format_desc(hint), model.index(0).data())
        self.assertIsInstance(model.index(0).data(Qt.DecorationRole), QIcon)
        self.assertIsInstance(model.index(0).data(Qt.ToolTipRole), str)



class TestUtils(GuiTest):
    def test_show_tip(self):
        w = QWidget()
        show_tip = IncreasingNumbersListValidator.show_tip
        show_tip(w, QPoint(100, 100), "Ha Ha")
        app = QApplication.instance()
        windows = app.topLevelWidgets()
        label = [tl for tl in windows
                 if tl.parent() is w and tl.objectName() == "tip-label"][0]
        self.assertTrue(label.isVisible())
        self.assertTrue(label.text() == "Ha Ha")

        show_tip(w, QPoint(100, 100), "Ha")
        self.assertTrue(label.text() == "Ha")
        show_tip(w, QPoint(100, 100), "")
        self.assertFalse(label.isVisible())

    def test_format_desc(self):
        self.assertEqual(format_desc(VarHint(Methods.MDL, ())),
                         Options[Methods.MDL].short_desc)
        self.assertEqual(format_desc(VarHint(Methods.EqualWidth, ("10", ))),
                         Options[Methods.EqualWidth].short_desc.format(10))
        self.assertEqual(format_desc(None),
                         Options[Methods.Default].short_desc)

        fwt = Methods.FixedWidthTime
        desc = Options[fwt].short_desc.format
        self.assertEqual(format_desc(VarHint(fwt, ("1", 0))), desc("1", "year"))
        self.assertEqual(format_desc(VarHint(fwt, ("2", 0))), desc("2", "years"))
        self.assertEqual(format_desc(VarHint(fwt, ("1", 2))), desc("1", "day"))
        self.assertEqual(format_desc(VarHint(fwt, ("2", 2))), desc("2", "days"))
        self.assertEqual(format_desc(VarHint(fwt, ("x", 2))), desc("x", "day(s)"))
        self.assertEqual(format_desc(VarHint(fwt, ("", 2))), desc("", "day(s)"))

    def test_fixed_width_disc(self):
        fw = partial(_fixed_width_discretization, None, None)
        for arg in ("", "5.3.1", "abc", "-5", "0"):
            self.assertIsInstance(fw(arg), str)

        with patch("Orange.preprocess.discretize.FixedWidth") as disc:
            self.assertNotIsInstance(fw("5.13"), str)
            disc.assert_called_with(5.13, 2)

            self.assertNotIsInstance(fw("5"), str)
            disc.assert_called_with(5, 0)

        with patch("Orange.preprocess.discretize.FixedWidth",
                   side_effect=TooManyIntervals):
            self.assertIsInstance(fw("42"), str)

    def test_fixed_time_width_disc(self):
        ftw = partial(_fixed_time_width_discretization, None, None)

        for arg in ("", "5.3.1", "5.3", "abc", "-5", "0"):
            self.assertIsInstance(ftw(arg, 1), str)

        with patch("Orange.preprocess.discretize.FixedTimeWidth") as disc:
            self.assertNotIsInstance(ftw("5", 2), str)
            disc.assert_called_with(5, 2)

            self.assertNotIsInstance(ftw("5", 3), str)
            disc.assert_called_with(35, 2)

            self.assertNotIsInstance(ftw("5", 4), str)
            disc.assert_called_with(5, 3)

        with patch("Orange.preprocess.discretize.FixedTimeWidth",
                   side_effect=TooManyIntervals):
            self.assertIsInstance(ftw("42", 3), str)

    def test_custom_discretization(self):
        cd = partial(_custom_discretization, None, None)

        for arg in ("", "4 5", "2, 1, 5", "1, foo, 13"):
            self.assertIsInstance(cd(arg), str)

        with patch("Orange.preprocess.discretize.Discretizer."
                   "create_discretized_var") as disc:
            cd("1, 1.25, 1.5, 4")
            disc.assert_called_with(None, [1, 1.25, 1.5, 4])

    def test_mdl_discretization(self):
        mdl = _mdl_discretization
        data = Table("iris")[::10]
        var = data.domain[0]
        with patch("Orange.preprocess.discretize.EntropyMDL") as mdldisc:
            mdl(data, var)
            mdldisc.return_value.assert_called_with(data, var)
            mdldisc.reset_mock()

            data = data[:, :4]
            self.assertIsInstance(mdl(data, var), str)
            mdldisc.assert_not_called()

            data = data.transform(Domain(data.domain[:3], data.domain[3]))
            self.assertIsInstance(mdl(data, var), str)
            mdldisc.assert_not_called()

    def test_var_key(self):
        self.assertEqual(variable_key(ContinuousVariable("foo")),
                         ("foo", False))
        self.assertEqual(variable_key(TimeVariable("bar")),
                         ("bar", True))


if __name__ == "__main__":
    unittest.main()



import numpy

from Orange.regression import Learner, Model
from Orange.statistics import distribution

__all__ = ["MeanLearner"]


class MeanLearner(Learner):
    """
    Fit a regression model that returns the average response (class) value.
    """
    def fit_storage(self, data):
        """
        Construct a :obj:`MeanModel` by computing the mean value of the given
        data.

        :param data: data table
        :type data: Orange.data.Table
        :return: regression model, which always returns mean value
        :rtype: :obj:`MeanModel`
        """
        if not data.domain.has_continuous_class:
            raise ValueError("regression.MeanLearner expects a domain with a "
                             "(single) numeric variable.")
        dist = distribution.get_distribution(data, data.domain.class_var)
        return MeanModel(dist)


# noinspection PyMissingConstructor
class MeanModel(Model):
    """
    A regression model that returns the average response (class) value.
    Instances can be constructed directly, by passing a distribution to the
    constructor, or by calling the :obj:`MeanLearner`.

    .. automethod:: __init__

    """
    def __init__(self, dist, domain=None):
        """
        Construct :obj:`Orange.regression.MeanModel` that always returns the
        mean value computed from the given distribution.

        If the distribution is empty, it constructs a model that returns zero.

        :param dist: domain for the `Table`
        :type dist: Orange.statistics.distribution.Continuous
        :return: regression model that returns mean value
        :rtype: :obj:`MeanModel`
        """
        # Don't call super().__init__ because it will raise an error since
        # domain is None.
        self.domain = domain
        self.dist = dist
        if dist.any():
            self.mean = self.dist.mean()
        else:
            self.mean = 0.0

    # noinspection PyPep8Naming
    def predict(self, X):
        """
        Return predictions (that is, the same mean value) for each given
        instance in `X`.

        :param X: data for which to make predictions
        :type X: :obj:`numpy.ndarray`
        :return: a vector of predictions
        :rtype: :obj:`numpy.ndarray`
        """
        return numpy.full(len(X), self.mean)

    def __str__(self):
        return 'MeanModel({})'.format(self.mean)

MeanLearner.__returns__ = MeanModel



# pylint: disable=protected-access
from unittest.mock import patch

import numpy as np
from sklearn.utils import check_random_state

from orangewidget.settings import Context

from Orange.data import Table, Domain, ContinuousVariable
from Orange.preprocess import Normalize
from Orange.widgets.tests.base import WidgetTest
from Orange.widgets.tests.utils import table_dense_sparse
from Orange.widgets.unsupervised.owlouvainclustering import OWLouvainClustering

# Deterministic tests
np.random.seed(42)


class TestOWLouvain(WidgetTest):
    def setUp(self):
        self.widget = self.create_widget(
            OWLouvainClustering, stored_settings={'auto_commit': False}
        )
        self.iris = Table('iris')[::5].copy()

    def tearDown(self):
        self.widget.onDeleteWidget()
        super().tearDown()

    def commit_and_wait(self, widget=None):
        widget = self.widget if widget is None else widget
        widget.commit()
        self.wait_until_stop_blocking(widget)

    def test_removing_data(self):
        self.send_signal(self.widget.Inputs.data, self.iris)
        self.commit_and_wait(self.widget)
        self.send_signal(self.widget.Inputs.data, None)
        self.commit_and_wait(self.widget)

    def test_clusters_ordered_by_size(self):
        """Cluster names should be sorted based on the number of instances."""
        x1 = np.array([[0, 0]] * 20)
        x2 = np.array([[1, 0]] * 15)
        x3 = np.array([[0, 1]] * 10)
        x4 = np.array([[1, 1]] * 5)
        data = np.vstack((x1, x2, x3, x4))
        # Remove any order depencence in data, not that this should affect it
        np.random.shuffle(data)

        table = Table.from_numpy(domain=Domain.from_numpy(X=data), X=data)

        self.send_signal(self.widget.Inputs.data, table)
        self.widget.k_neighbors = 4
        self.commit_and_wait()
        output = self.get_output(self.widget.Outputs.annotated_data)

        clustering = output.get_column('Cluster').astype(int)
        counts = np.bincount(clustering)
        np.testing.assert_equal(counts, sorted(counts, reverse=True))

    def test_empty_dataset(self):
        # Prepare a table with 5 rows with only meta attributes
        meta = np.array([0] * 5)
        meta_var = ContinuousVariable(name='meta_var')
        table = Table.from_domain(domain=Domain([], metas=[meta_var]), n_rows=5)
        with table.unlocked():
            table.set_column(meta_var, meta)

        self.send_signal(self.widget.Inputs.data, table)
        self.commit_and_wait()
        self.assertTrue(self.widget.Error.empty_dataset.is_shown())

    def test_do_not_recluster_on_same_data(self):
        """Do not recluster data points when targets or metas change."""

        # Prepare some dummy data
        x = np.eye(5)
        y1, y2 = np.ones((5, 1)), np.ones((5, 2))
        meta1, meta2 = np.ones((5, 1)), np.ones((5, 2))

        table1 = Table.from_numpy(
            domain=Domain.from_numpy(X=x, Y=y1, metas=meta1),
            X=x, Y=y1, metas=meta1,
        )
        # X is same, should not cause update
        table2 = Table.from_numpy(
            domain=Domain.from_numpy(X=x, Y=y2, metas=meta2),
            X=x, Y=y2, metas=meta2,
        )
        # X is different, should cause update
        table3 = table1.copy()
        with table3.unlocked():
            table3.X[:, 0] = 1

        with patch.object(self.widget, '_invalidate_output') as commit:
            self.send_signal(self.widget.Inputs.data, table1)
            self.commit_and_wait()
            call_count = commit.call_count

            # Sending data with same X should not recompute the clustering
            self.send_signal(self.widget.Inputs.data, table2)
            self.commit_and_wait()
            self.assertEqual(call_count, commit.call_count)

            # Sending data with different X should recompute the clustering
            self.send_signal(self.widget.Inputs.data, table3)
            self.commit_and_wait()
            self.assertEqual(call_count + 1, commit.call_count)

    def test_only_recluster_when_necessary_pca_components_change(self):
        # Compute clustering on some data
        self.send_signal(self.widget.Inputs.data, self.iris)

        # When PCA checkbox is ticked, any update to slider should invalidate results
        self.widget.apply_pca_cbx.setChecked(True)
        self.widget.pca_components_slider.setValue(2)
        self.commit_and_wait()

        with patch.object(self.widget, '_invalidate_output') as invalidate:
            # Change slider value, this should invalidate output
            self.widget.pca_components_slider.setValue(4)
            self.commit_and_wait()
            self.assertEqual(invalidate.call_count, 1)

        with patch.object(self.widget, '_invalidate_output') as invalidate:
            # Don't change slider value, this shouldn't do anything
            self.widget.pca_components_slider.setValue(4)
            self.commit_and_wait()
            invalidate.assert_not_called()

        # When PCA checkbox is not ticked updating the slider should have no effect
        self.widget.apply_pca_cbx.setChecked(False)
        self.widget.pca_components_slider.setValue(2)
        self.commit_and_wait()

        with patch.object(self.widget, '_invalidate_output') as invalidate:
            # Change slider value, this should invalidate output
            self.widget.pca_components_slider.setValue(4)
            self.commit_and_wait()
            invalidate.assert_not_called()

    def test_invalidate(self):
        # pylint: disable=protected-access
        data = self.iris
        self.send_signal(self.widget.Inputs.data, data)
        self.widget.commit()
        out = self.get_output(self.widget.Outputs.annotated_data)
        self.assertIsNotNone(out)
        # invalidate the partitioning
        p = self.widget.partition
        g = self.widget.graph
        self.widget._invalidate_partition()
        # not in auto commit mode.
        self.assertTrue(self.widget.Information.modified.is_shown())
        self.widget.commit()
        out = self.get_output(self.widget.Outputs.annotated_data)
        self.assertIsNotNone(out)
        self.assertIs(self.widget.graph, g)
        self.assertIsNot(self.widget.partition, p)

        self.assertFalse(self.widget.Information.modified.is_shown())
        self.widget._invalidate_graph()
        self.assertIsNone(self.widget.graph)
        self.assertIsNotNone(self.widget.pca_projection)
        self.assertTrue(self.widget.Information.modified.is_shown())
        self.widget._invalidate_pca_projection()
        self.assertIsNone(self.widget.pca_projection)
        self.widget.commit()
        self.get_output(self.widget.Outputs.annotated_data)
        self.assertFalse(self.widget.Information.modified.is_shown())

    def test_deterministic_clustering(self):
        # Compute clustering on iris
        self.send_signal(self.widget.Inputs.data, self.iris)
        self.commit_and_wait()
        result1 = self.get_output(self.widget.Outputs.annotated_data)

        # Reset widget state
        self.send_signal(self.widget.Inputs.data, None)

        # Compute clustering on iris again
        self.send_signal(self.widget.Inputs.data, self.iris)
        self.commit_and_wait()
        result2 = self.get_output(self.widget.Outputs.annotated_data)

        # Ensure that clustering was the same in both instances
        np.testing.assert_equal(result1.metas, result2.metas)

    @table_dense_sparse
    def test_normalize_data(self, prepare_table):
        """Check that normalization is called at the proper times."""
        data = prepare_table(self.iris)

        # Enable checkbox
        self.widget.controls.normalize.setChecked(True)
        self.assertTrue(self.widget.controls.normalize.isChecked())
        with patch("Orange.preprocess.Normalize", wraps=Normalize) as normalize:
            self.send_signal(self.widget.Inputs.data, data)
            self.wait_until_stop_blocking()
            self.assertTrue(self.widget.controls.normalize.isEnabled())
            normalize.assert_called_once()

        # Disable checkbox
        self.widget.controls.normalize.setChecked(False)
        self.assertFalse(self.widget.controls.normalize.isChecked())
        with patch("Orange.preprocess.Normalize", wraps=Normalize) as normalize:
            self.send_signal(self.widget.Inputs.data, data)
            self.wait_until_stop_blocking()
            self.assertTrue(self.widget.controls.normalize.isEnabled())
            normalize.assert_not_called()

    def test_dense_and_sparse_return_same_result(self):
        """Check that Louvain clustering returns identical results for both
        dense and sparse data."""
        random_state = check_random_state(42)

        # Randomly set some values to zero
        dense_data = self.iris
        mask = random_state.beta(1, 2, size=self.iris.X.shape) > 0.5
        with dense_data.unlocked():
            dense_data.X[mask] = 0
        sparse_data = dense_data.to_sparse()

        def _compute_clustering(data):
            self.send_signal(self.widget.Inputs.data, data)
            self.wait_until_stop_blocking()
            result = self.get_output(self.widget.Outputs.annotated_data)
            self.send_signal(self.widget.Inputs.data, None)
            return result

        # Disable normalization
        self.widget.controls.normalize.setChecked(False)
        dense_result = _compute_clustering(dense_data)
        sparse_result = _compute_clustering(sparse_data)
        np.testing.assert_equal(dense_result.metas, sparse_result.metas)

        # Enable normalization
        self.widget.controls.normalize.setChecked(True)
        dense_result = _compute_clustering(dense_data)
        sparse_result = _compute_clustering(sparse_data)
        np.testing.assert_equal(dense_result.metas, sparse_result.metas)

    def test_graph_output(self):
        w = self.widget

        # This test executes only if network add-on is installed
        if not hasattr(w.Outputs, "graph"):
            return

        self.send_signal(w.Inputs.data, self.iris)
        graph = self.get_output(w.Outputs.graph)
        self.assertEqual(len(graph.nodes), len(self.iris))

        self.send_signal(w.Inputs.data, None)
        graph = self.get_output(w.Outputs.graph)
        self.assertIsNone(graph)

    def test_migrate_settings(self):
        # any context settings are removed
        settings = {"context_settings": []}
        self.widget.migrate_settings(settings, 1)
        self.assertEqual(len(settings), 0)

        # context settings become ordinary settings
        settings = {"context_settings": [Context(values={'__version__': 1,
                                                         'apply_pca': (True, -2),
                                                         'k_neighbors': (29, -2),
                                                         'metric_idx': (1, -2),
                                                         'normalize': (False, -2),
                                                         'pca_components': (10, -2),
                                                         'resolution': (1.0, -2)})]}
        self.widget.migrate_settings(settings, 1)
        correct = {'apply_pca': True, 'k_neighbors': 29, 'metric_idx': 1,
                   'normalize': False, 'pca_components': 10, 'resolution': 1.0}
        self.assertEqual(sorted(settings.items()), sorted(correct.items()))



#pylint: disable=unsubscriptable-object
import datetime
import unittest
from collections import namedtuple
from functools import partial
from itertools import chain
from typing import List

import numpy as np
from AnyQt.QtCore import QItemSelection, QItemSelectionRange, \
    QItemSelectionModel, Qt

from orangewidget.settings import Context

from Orange.data import Table, Domain, StringVariable, ContinuousVariable, \
    DiscreteVariable, TimeVariable
from Orange.widgets.tests.base import WidgetTest, datasets
from Orange.widgets.tests.utils import simulate, table_dense_sparse
from Orange.widgets.data.owfeaturestatistics import \
    OWFeatureStatistics

VarDataPair = namedtuple('VarDataPair', ['variable', 'data'])

# Continuous variable variations
continuous_full = VarDataPair(
    ContinuousVariable('continuous_full'),
    np.array([0, 1, 2, 3, 4], dtype=float),
)
continuous_missing = VarDataPair(
    ContinuousVariable('continuous_missing'),
    np.array([0, 1, 2, np.nan, 4], dtype=float),
)
continuous_all_missing = VarDataPair(
    ContinuousVariable('continuous_all_missing'),
    np.array([np.nan] * 5, dtype=float),
)
continuous_same = VarDataPair(
    ContinuousVariable('continuous_same'),
    np.array([3] * 5, dtype=float),
)
continuous = [
    continuous_full, continuous_missing, continuous_all_missing,
    continuous_same
]

# Unordered discrete variable variations
rgb_full = VarDataPair(
    DiscreteVariable('rgb_full', values=('r', 'g', 'b')),
    np.array([0, 1, 1, 1, 2], dtype=float),
)
rgb_missing = VarDataPair(
    DiscreteVariable('rgb_missing', values=('r', 'g', 'b')),
    np.array([0, 1, 1, np.nan, 2], dtype=float),
)
rgb_all_missing = VarDataPair(
    DiscreteVariable('rgb_all_missing', values=('r', 'g', 'b')),
    np.array([np.nan] * 5, dtype=float),
)
rgb_bins_missing = VarDataPair(
    DiscreteVariable('rgb_bins_missing', values=('r', 'g', 'b')),
    np.array([np.nan, 1, 1, 1, np.nan], dtype=float),
)
rgb_same = VarDataPair(
    DiscreteVariable('rgb_same', values=('r', 'g', 'b')),
    np.array([2] * 5, dtype=float),
)
rgb = [
    rgb_full, rgb_missing, rgb_all_missing, rgb_bins_missing, rgb_same
]

# Ordered discrete variable variations
ints_full = VarDataPair(
    DiscreteVariable('ints_full', values=('2', '3', '4')),
    np.array([0, 1, 1, 1, 2], dtype=float),
)
ints_missing = VarDataPair(
    DiscreteVariable('ints_missing', values=('2', '3', '4')),
    np.array([0, 1, 1, np.nan, 2], dtype=float),
)
ints_all_missing = VarDataPair(
    DiscreteVariable('ints_all_missing', values=('2', '3', '4')),
    np.array([np.nan] * 5, dtype=float),
)
ints_bins_missing = VarDataPair(
    DiscreteVariable('ints_bins_missing', values=('2', '3', '4')),
    np.array([np.nan, 1, 1, 1, np.nan], dtype=float),
)
ints_same = VarDataPair(
    DiscreteVariable('ints_same', values=('2', '3', '4')),
    np.array([0] * 5, dtype=float),
)
ints = [
    ints_full, ints_missing, ints_all_missing, ints_bins_missing, ints_same
]

discrete = list(chain(rgb, ints))


def _to_timestamps(years):
    return [datetime.datetime(year, 1, 1).timestamp() if not np.isnan(year)
            else np.nan for year in years]


# Time variable variations, windows timestamps need to be valid timestamps so
# we'll just fill it in with arbitrary years
time_full = VarDataPair(
    TimeVariable('time_full', have_date=True, have_time=True),
    np.array(_to_timestamps([2000, 2001, 2002, 2003, 2004]), dtype=float),
)
time_missing = VarDataPair(
    TimeVariable('time_missing'),
    np.array(_to_timestamps([2000, np.nan, 2001, 2003, 2004]), dtype=float),
)
time_all_missing = VarDataPair(
    TimeVariable('time_all_missing'),
    np.array(_to_timestamps([np.nan] * 5), dtype=float),
)
time_same = VarDataPair(
    TimeVariable('time_same', have_date=True, have_time=True),
    np.array(_to_timestamps([2004] * 5), dtype=float),
)
time_negative = VarDataPair(
    TimeVariable('time_negative', have_date=True, have_time=True),
    np.array([0, -1, 24 * 60 * 60], dtype=float),
)
time = [
    time_full, time_missing, time_all_missing, time_same, time_negative
]

# String variable variations
string_full = VarDataPair(
    StringVariable('string_full'),
    np.array(['a', 'b', 'c', 'd', 'e'], dtype=object),
)
string_missing = VarDataPair(
    StringVariable('string_missing'),
    np.array(['a', 'b', 'c', StringVariable.Unknown, 'e'], dtype=object),
)
string_all_missing = VarDataPair(
    StringVariable('string_all_missing'),
    np.array([StringVariable.Unknown] * 5, dtype=object),
)
string_same = VarDataPair(
    StringVariable('string_same'),
    np.array(['a'] * 5, dtype=object),
)
string = [
    string_full, string_missing, string_all_missing, string_same
]


def make_table(attributes, target=None, metas=None):
    """Build an instance of a table given various variables.

    Parameters
    ----------
    attributes : Iterable[Tuple[Variable, np.array]
    target : Optional[Iterable[Tuple[Variable, np.array]]
    metas : Optional[Iterable[Tuple[Variable, np.array]]

    Returns
    -------
    Table

    """
    attribute_vars, attribute_vals = list(zip(*attributes))
    attribute_vals = np.array(attribute_vals).T

    target_vars, target_vals = None, None
    if target is not None:
        target_vars, target_vals = list(zip(*target))
        target_vals = np.array(target_vals).T

    meta_vars, meta_vals = None, None
    if metas is not None:
        meta_vars, meta_vals = list(zip(*metas))
        meta_vals = np.array(meta_vals).T

    return Table.from_numpy(
        Domain(attribute_vars, class_vars=target_vars, metas=meta_vars),
        X=attribute_vals, Y=target_vals, metas=meta_vals,
    )


class TestVariousDataSets(WidgetTest):
    def setUp(self):
        self.widget = self.create_widget(
            OWFeatureStatistics, stored_settings={'auto_commit': False}
        )

    def force_render_table(self):
        """Some fields e.g. histograms are only initialized when they actually
        need to be rendered"""
        model = self.widget.model
        for i in range(model.rowCount()):
            for j in range(model.columnCount()):
                model.data(model.index(i, j), Qt.DisplayRole)

    def run_through_variables(self):
        simulate.combobox_run_through_all(
            self.widget.cb_color_var, callback=self.force_render_table)

    @table_dense_sparse
    def test_runs_on_iris(self, prepare_table):
        self.send_signal(self.widget.Inputs.data, prepare_table(Table('iris')))

    def test_does_not_crash_on_data_removal(self):
        self.send_signal(self.widget.Inputs.data, make_table(discrete))
        self.send_signal(self.widget.Inputs.data, None)

    def test_does_not_crash_on_empty_domain(self):
        empty_data = Table('iris').transform(Domain([]))
        self.send_signal(self.widget.Inputs.data, empty_data)

    # No missing values
    @table_dense_sparse
    def test_on_data_with_no_missing_values(self, prepare_table):
        data = make_table([continuous_full, rgb_full, ints_full, time_full])
        self.send_signal(self.widget.Inputs.data, prepare_table(data))
        self.run_through_variables()

    @table_dense_sparse
    def test_on_data_with_no_missing_values_full_domain(self, prepare_table):
        data = make_table([continuous_full, time_full], [ints_full], [rgb_full])
        self.send_signal(self.widget.Inputs.data, prepare_table(data))
        self.run_through_variables()

    # With missing values
    @table_dense_sparse
    def test_on_data_with_missing_continuous_values(self, prepare_table):
        data = make_table([continuous_full, continuous_missing, rgb_full, ints_full, time_full])
        self.send_signal(self.widget.Inputs.data, prepare_table(data))
        self.run_through_variables()

    @table_dense_sparse
    def test_on_data_with_missing_discrete_values(self, prepare_table):
        data = make_table([continuous_full, rgb_full, rgb_missing, ints_full, time_full])
        self.send_signal(self.widget.Inputs.data, prepare_table(data))
        self.run_through_variables()

    @table_dense_sparse
    def test_on_data_with_discrete_values_all_the_same(self, prepare_table):
        data = make_table([continuous_full], [ints_same, rgb_same])
        self.send_signal(self.widget.Inputs.data, prepare_table(data))
        self.run_through_variables()

    @table_dense_sparse
    def test_on_data_with_continuous_values_all_the_same(self, prepare_table):
        data = make_table([ints_full, ints_same], [continuous_same, continuous_full])
        self.send_signal(self.widget.Inputs.data, prepare_table(data))
        self.run_through_variables()

    @table_dense_sparse
    def test_on_data_with_negative_timestamps(self, prepare_table):
        data = make_table([time_negative])
        self.send_signal(self.widget.Inputs.data, prepare_table(data))
        self.run_through_variables()

    def test_switching_to_dataset_with_no_target_var(self):
        """Switching from data set with target variable to a data set with
        no target variable should not result in crash."""
        data1 = make_table([continuous_full, ints_full], [ints_same, rgb_same])
        data2 = make_table([rgb_full, ints_full])

        self.send_signal(self.widget.Inputs.data, data1)
        self.force_render_table()

        self.send_signal(self.widget.Inputs.data, data2)
        self.force_render_table()

    def test_switching_to_dataset_with_target_var(self):
        """Switching from data set with no target variable to a data set with
        a target variable should not result in crash."""
        data1 = make_table([rgb_full, ints_full])
        data2 = make_table([continuous_full, ints_full], [ints_same, rgb_same])

        self.send_signal(self.widget.Inputs.data, data1)
        self.force_render_table()

        self.send_signal(self.widget.Inputs.data, data2)
        self.force_render_table()

    def test_on_edge_case_datasets(self):
        for data in datasets.datasets():
            try:
                self.send_signal(self.widget.Inputs.data, data)
                self.force_render_table()
            except Exception as e:
                raise AssertionError(f"Failed on `{data.name}`") from e

    def test_header_resize_aspect_ratio(self):
        self.widget.show()  # must be visible for header resize to work
        size = self.widget.size()
        self.widget.resize(size.width() + 2000, size.height())
        self.assertEqual(
            self.widget.table_view.verticalHeader().defaultSectionSize(),
            self.widget.table_view.MAXIMUM_HISTOGRAM_HEIGHT
        )


def select_rows(rows: List[int], widget: OWFeatureStatistics):
    """Since the widget sorts the rows, selecting rows isn't trivial."""
    indices = widget.model.mapToSourceRows(rows)

    selection = QItemSelection()
    for idx in indices:
        selection.append(QItemSelectionRange(
            widget.model.index(idx, 0),
            widget.model.index(idx, widget.model.columnCount() - 1)
        ))

    widget.table_view.selectionModel().select(
        selection, QItemSelectionModel.ClearAndSelect)


class TestFeatureStatisticsOutputs(WidgetTest):
    def setUp(self):
        self.widget = self.create_widget(
            OWFeatureStatistics, stored_settings={'auto_commit': False}
        )
        self.data = make_table(
            [continuous_full, continuous_missing],
            target=[rgb_full, rgb_missing], metas=[ints_full, ints_missing]
        )
        self.send_signal(self.widget.Inputs.data, self.data)
        self.select_rows = partial(select_rows, widget=self.widget)

    def test_changing_data_updates_output(self):
        # Test behaviour of widget when auto commit is OFF
        self.widget.auto_commit = False

        # We start of with some data and select some rows
        self.send_signal(self.widget.Inputs.data, self.data)
        self.select_rows([0])
        # By default, nothing should be sent since auto commit is off
        self.assertIsNone(self.get_output(self.widget.Outputs.reduced_data))
        # When we commit, the data should be on the output
        self.widget.commit.now()
        self.assertIsNotNone(self.get_output(self.widget.Outputs.reduced_data))

        # Send some new data
        iris = Table('iris')
        self.send_signal(self.widget.Inputs.data, iris)
        # By default, there should be nothing on the output
        self.assertIsNone(self.get_output(self.widget.Outputs.reduced_data))
        # Nothing should change after commit, since we haven't selected any rows
        self.widget.commit.now()
        self.assertIsNone(self.get_output(self.widget.Outputs.reduced_data))

    def test_changing_data_updates_output_with_autocommit(self):
        # Test behaviour of widget when auto commit is ON
        self.widget.auto_commit = True

        # We start of with some data and select some rows
        self.send_signal(self.widget.Inputs.data, self.data)
        self.select_rows([0])
        # Selecting rows should send data to output
        self.assertIsNotNone(self.get_output(self.widget.Outputs.reduced_data))

        # Send some new data
        iris = Table('iris')
        self.send_signal(self.widget.Inputs.data, iris)
        # Don't select anything, so the outputs should be empty
        self.assertIsNone(self.get_output(self.widget.Outputs.reduced_data))

        # Now let's switch back to the original data, where we had selected row 0,
        # we expect that to be sent to output
        self.send_signal(self.widget.Inputs.data, self.data)
        self.assertIsNotNone(self.get_output(self.widget.Outputs.reduced_data))

    def test_sends_single_attribute_table_to_output(self):
        # Check if selecting a single attribute row
        self.select_rows([0])
        self.widget.commit.now()

        desired_domain = Domain(attributes=[continuous_full.variable])
        output = self.get_output(self.widget.Outputs.reduced_data)
        self.assertEqual(output.domain, desired_domain)

    def test_sends_multiple_attribute_table_to_output(self):
        # Check if selecting a single attribute row
        self.select_rows([0, 1])
        self.widget.commit.now()

        desired_domain = Domain(attributes=[
            continuous_full.variable, continuous_missing.variable,
        ])
        output = self.get_output(self.widget.Outputs.reduced_data)
        self.assertEqual(output.domain, desired_domain)

    def test_sends_single_class_var_table_to_output(self):
        self.select_rows([2])
        self.widget.commit.now()

        desired_domain = Domain(attributes=[], class_vars=[rgb_full.variable])
        output = self.get_output(self.widget.Outputs.reduced_data)
        self.assertEqual(output.domain, desired_domain)

    def test_sends_single_meta_table_to_output(self):
        self.select_rows([4])
        self.widget.commit.now()

        desired_domain = Domain(attributes=[], metas=[ints_full.variable])
        output = self.get_output(self.widget.Outputs.reduced_data)
        self.assertEqual(output.domain, desired_domain)

    def test_sends_multiple_var_types_table_to_output(self):
        self.select_rows([0, 2, 4])
        self.widget.commit.now()

        desired_domain = Domain(
            attributes=[continuous_full.variable],
            class_vars=[rgb_full.variable],
            metas=[ints_full.variable],
        )
        output = self.get_output(self.widget.Outputs.reduced_data)
        self.assertEqual(output.domain, desired_domain)

    def test_sends_all_samples_to_output(self):
        """All rows should be sent to output for selected column."""
        self.select_rows([0, 2])
        self.widget.commit.now()

        selected_vars = Domain(
            attributes=[continuous_full.variable],
            class_vars=[rgb_full.variable],
        )

        output = self.get_output(self.widget.Outputs.reduced_data)
        np.testing.assert_equal(output.X, self.data[:, selected_vars.variables].X)
        np.testing.assert_equal(output.Y, self.data[:, selected_vars.variables].Y)

    def test_clearing_selection_sends_none_to_output(self):
        """Clearing all the selected rows should send `None` to output."""
        self.select_rows([0])
        self.widget.commit.now()
        self.assertIsNotNone(self.get_output(self.widget.Outputs.reduced_data))

        self.widget.table_view.clearSelection()
        self.widget.commit.now()
        self.assertIsNone(self.get_output(self.widget.Outputs.reduced_data))

    def test_output_statistics(self):
        self.widget.auto_commit = True

        data = make_table([continuous_full, continuous_missing,
                           rgb_full, rgb_missing])
        self.send_signal(self.widget.Inputs.data, data)
        output = self.get_output(self.widget.Outputs.statistics)
        np.testing.assert_almost_equal(
            output.X,
            [[2, 2, 0.7071068, 0, 4, 0],
             [1.75, 1.5, 0.8451543, 0, 4, 1],
             [np.nan, np.nan, 0.9502705, np.nan, np.nan, 0],
             [np.nan, np.nan, 1.0397208, np.nan, np.nan, 1]],
        )
        np.testing.assert_equal(
            output.metas,
            [["continuous_full", "0"],
             ["continuous_missing", "0"],
             ["rgb_full", "g"],
             ["rgb_missing", "g"]])

        data = make_table([continuous_full, continuous_missing])
        self.send_signal(self.widget.Inputs.data, data)
        output = self.get_output(self.widget.Outputs.statistics)
        np.testing.assert_almost_equal(
            output.X,
            [[2, 2, 0.7071068, 0, 4, 0],
             [1.75, 1.5, 0.8451543, 0, 4, 1]]
        )
        np.testing.assert_equal(
            output.metas,
            [["continuous_full", "0"],
             ["continuous_missing", "0"]])

        data = make_table([rgb_full, rgb_missing])
        self.send_signal(self.widget.Inputs.data, data)
        output = self.get_output(self.widget.Outputs.statistics)
        np.testing.assert_almost_equal(
            output.X,
            [[0.9502705, 0],
             [1.0397208, 1]],
        )
        np.testing.assert_equal(
            output.metas,
            [["rgb_full", "g"],
             ["rgb_missing", "g"]])

        self.send_signal(self.widget.Inputs.data, None)
        output = self.get_output(self.widget.Outputs.statistics)
        self.assertIsNone(output)

    def test_output_combinations(self):
        # No selection -> reduced_data is not output, statistics is present
        self.widget.commit.now()
        self.assertIsNone(self.get_output(self.widget.Outputs.reduced_data))
        self.assertEqual(len(self.get_output(self.widget.Outputs.statistics)),
                         self.widget.model.rowCount())

        # Has selection -> all outputs present
        self.select_rows([0, 1])
        self.widget.commit.now()
        outp = self.get_output(self.widget.Outputs.reduced_data)
        self.assertEqual(len(outp), len(self.data))
        self.assertEqual(len(outp.domain.variables), 2)
        self.assertEqual(len(self.get_output(self.widget.Outputs.statistics)),
                         self.widget.model.rowCount())

        # No data -> no output
        self.send_signal(self.widget.Inputs.data, None)
        self.assertIsNone(self.get_output(self.widget.Outputs.reduced_data))
        self.assertIsNone(self.get_output(self.widget.Outputs.reduced_data))


class TestFeatureStatisticsUI(WidgetTest):
    def setUp(self):
        self.widget = self.create_widget(
            OWFeatureStatistics, stored_settings={'auto_commit': False}
        )
        self.data1 = Table('iris')
        self.data2 = Table('zoo')
        self.select_rows = partial(select_rows, widget=self.widget)

    def test_restores_previous_selection(self):
        """Widget should remember selection with domain context handler."""
        # Send data and select rows
        domain1 = self.data1.domain
        self.send_signal(self.widget.Inputs.data, self.data1)
        self.select_rows([0, 2])
        self.assertEqual(set(self.widget.selected_vars),
                         {domain1[0], domain1[2]})

        # Sending new data clears selection
        self.send_signal(self.widget.Inputs.data, self.data2)
        self.assertEqual(len(self.widget.selected_vars), 0)

        # Sending back the old data restores the selection
        iris3 = self.data1.transform(
            Domain([domain1[2], domain1[0], domain1[1]], domain1.class_var))
        self.send_signal(self.widget.Inputs.data, iris3)
        self.assertEqual(set(self.widget.selected_vars),
                         {domain1[0], domain1[2]})

    def test_settings_migration_to_ver21(self):
        settings = {
            'controlAreaVisible': True, 'savedWidgetGeometry': '',
            '__version__': 1,
            'context_settings': [
                Context(
                    values={'auto_commit': (True, -2),
                            'color_var': ('iris', 101),
                            'selected_rows': [1, 4],
                            'sorting': ((1, 0), -2), '__version__': 1},
                    attributes={'petal length': 2, 'petal width': 2,
                                'sepal length': 2, 'sepal width': 2},
                    metas={'iris': 1})]
        }
        widget = self.create_widget(OWFeatureStatistics,
                                    stored_settings=settings)
        self.send_signal(widget.Inputs.data, self.data1)
        domain = self.data1.domain
        self.assertEqual(widget.selected_vars, [domain["petal width"],
                                                domain["iris"]])

    def test_report(self):
        self.send_signal(self.widget.Inputs.data, self.data1)

        self.widget.report_button.click()
        report_text = self.widget.report_html

        self.assertIn("<table>", report_text)
        self.assertEqual(6, report_text.count("<tr>"))  # header + 5 rows


class TestSummary(WidgetTest):
    def setUp(self):
        self.widget = self.create_widget(OWFeatureStatistics)
        self.data = make_table(
            [continuous_full, continuous_missing],
            target=[rgb_full, rgb_missing], metas=[ints_full, ints_missing]
        )
        self.select_rows = partial(select_rows, widget=self.widget)


if __name__ == "__main__":
    unittest.main()



import unittest

import numpy as np

from Orange.data import \
    Domain, Table, \
    DiscreteVariable, ContinuousVariable, TimeVariable, StringVariable
from Orange.preprocess.impute import ReplaceUnknownsRandom, ReplaceUnknowns, \
    FixedValueByType, ReplaceUnknownsModel
from Orange.regression import LinearRegressionLearner
from Orange.statistics.distribution import Discrete


class TestReplaceUnknowns(unittest.TestCase):
    def test_equality(self):
        v1 = ContinuousVariable("x")
        v2 = ContinuousVariable("x")
        v3 = ContinuousVariable("y")

        t1 = ReplaceUnknowns(v1, 0)
        t1a = ReplaceUnknowns(v2, 0)
        t2 = ReplaceUnknowns(v3, 0)
        self.assertEqual(t1, t1)
        self.assertEqual(t1, t1a)
        self.assertNotEqual(t1, t2)

        self.assertEqual(hash(t1), hash(t1a))
        self.assertNotEqual(hash(t1), hash(t2))

        t1 = ReplaceUnknowns(v1, 0)
        t1a = ReplaceUnknowns(v1, 1)
        self.assertNotEqual(t1, t1a)
        self.assertNotEqual(hash(t1), hash(t1a))


class TestReplaceUnknownsRandom(unittest.TestCase):
    def test_equality(self):
        v1 = DiscreteVariable("x", tuple("abc"))
        v2 = DiscreteVariable("x", tuple("abc"))
        v3 = DiscreteVariable("y", tuple("abc"))

        d1 = Discrete([1, 2, 3], v1)
        d2 = Discrete([1, 2, 3], v2)
        d3 = Discrete([1, 2, 3], v3)

        t1 = ReplaceUnknownsRandom(v1, d1)
        t1a = ReplaceUnknownsRandom(v2, d2)
        t2 = ReplaceUnknownsRandom(v3, d3)
        self.assertEqual(t1, t1)
        self.assertEqual(t1, t1a)
        self.assertNotEqual(t1, t2)

        self.assertEqual(hash(t1), hash(t1a))
        self.assertNotEqual(hash(t1), hash(t2))

        d1[1] += 1
        self.assertNotEqual(t1, t1a)
        self.assertNotEqual(hash(t1), hash(t1a))


class TestFixedValuesByType(unittest.TestCase):
    def setUp(self):
        domain = Domain(
            [DiscreteVariable("d", values=tuple("abc")),
             ContinuousVariable("c"),
             TimeVariable("t")],
            [],
            [StringVariable("s")]
        )
        n = np.nan
        self.data = Table(
            domain,
            np.array([[1, n, 15], [n, 42, n]]),
            np.empty((2, 0)),
            np.array([["foo"], [""]]))

    def test_none_defined(self):
        d, c, t = self.data.domain.attributes
        s,  = self.data.domain.metas

        imputer = FixedValueByType()
        for var in (d, c, t):
            imp = imputer(self.data, var)
            self.assertIsInstance(imp.compute_value, ReplaceUnknowns)
            self.assertTrue(np.isnan(imp.compute_value.value))
        imp = imputer(self.data, s)
        self.assertIsInstance(imp.compute_value, ReplaceUnknowns)
        self.assertIsNone(imp.compute_value.value)

    def test_all_defined(self):
        d, c, t = self.data.domain.attributes
        s, = self.data.domain.metas

        imputer = FixedValueByType(
            default_discrete=1, default_continuous=42,
            default_string="foo", default_time=3.14)

        self.assertEqual(imputer(self.data, d).compute_value.value, 1)
        self.assertEqual(imputer(self.data, c).compute_value.value, 42)
        self.assertEqual(imputer(self.data, t).compute_value.value, 3.14)
        self.assertEqual(imputer(self.data, s).compute_value.value, "foo")

    def test_with_default(self):
        s, = self.data.domain.metas

        imputer = FixedValueByType(
            default_discrete=1, default_continuous=42,
            default_string="foo", default_time=3.14)

        self.assertEqual(
            imputer(self.data, s, default="bar").compute_value.value,
            "bar")


class TestReplaceUnknownsModel(unittest.TestCase):
    def test_eq(self):
        iris = Table("iris")

        v1 = iris.domain[0]
        v2 = iris.domain[0]
        v3 = iris.domain[1]

        l = LinearRegressionLearner()
        def new_target(t):
            dom = Domain(iris.domain[2:], class_vars=[t])
            return iris.transform(dom)

        mod1 = l(new_target(v1))
        t1 = ReplaceUnknownsModel(v1, mod1)
        t1a = ReplaceUnknownsModel(v2, mod1)
        t2 = ReplaceUnknownsModel(v3, l(new_target(v3)))

        self.assertEqual(t1, t1)
        self.assertEqual(t1, t1a)
        self.assertNotEqual(t1, t2)

        # the following should be equal, but will not be unless __eq__ for that
        # particular model is defined
        t1b = ReplaceUnknownsModel(v1, l(new_target(v1)))
        self.assertNotEqual(t1, t1b)  # this is WRONG

        self.assertEqual(hash(t1), hash(t1a))
        self.assertNotEqual(hash(t1), hash(t2))


if __name__ == "__main__":
    unittest.main()



import unittest
from unittest.mock import patch

import numpy as np
from AnyQt.QtCore import QPointF

from Orange.widgets.visualize.utils.plotutils import InteractiveViewBox


class TestInteractiveViewBox(unittest.TestCase):

    def test_update_scale_box(self):
        view_box = InteractiveViewBox(graph=None)
        with patch.object(view_box, "mapToView", lambda x: x):
            view_box.updateScaleBox(QPointF(0, 0), QPointF(2, 2))
            tr = view_box.rbScaleBox.transform()
            self.assertFalse(tr.isRotating())
            trm = [[tr.m11(), tr.m12(), tr.m13()],
                   [tr.m21(), tr.m22(), tr.m23()],
                   [tr.m31(), tr.m32(), tr.m33()]]
            np.testing.assert_equal(trm, [[2, 0, 0], [0, 2, 0], [0, 0, 1]])


if __name__ == '__main__':
    unittest.main()



import unittest
import warnings
from unittest.mock import Mock

import numpy as np

from Orange.data.util import scale, one_hot, SharedComputeValue
import Orange

class TestDataUtil(unittest.TestCase):
    def test_scale(self):
        np.testing.assert_equal(scale([0, 1, 2], -1, 1), [-1, 0, 1])
        np.testing.assert_equal(scale([3, 3, 3]), [1, 1, 1])
        np.testing.assert_equal(scale([.1, .5, np.nan]), [0, 1, np.nan])
        np.testing.assert_equal(scale(np.array([])), np.array([]))

    def test_one_hot(self):
        np.testing.assert_equal(
            one_hot([0, 1, 2, 1], int), [[1, 0, 0],
                                         [0, 1, 0],
                                         [0, 0, 1],
                                         [0, 1, 0]])
        np.testing.assert_equal(one_hot([], int), np.zeros((0, 0), dtype=int))


class DummyPlus(SharedComputeValue):

    def compute(self, data, shared_data):
        return data.X[:, 0] + shared_data


class DummyTable(Orange.data.Table):
    pass


class TestSharedComputeValue(unittest.TestCase):

    def test_compat_compute_value(self):
        data = Orange.data.Table("iris")
        obj = DummyPlus(lambda data: 1.)
        res = obj(data)
        obj = lambda data: data.X[:, 0] + 1.
        res2 = obj(data)
        np.testing.assert_equal(res, res2)

    def test_with_row_indices(self):
        obj = DummyPlus(lambda data: 1.)
        data = Orange.data.Table("iris")
        domain = Orange.data.Domain([Orange.data.ContinuousVariable("cv", compute_value=obj)])
        data1 = Orange.data.Table.from_table(domain, data)[:10]
        data2 = Orange.data.Table.from_table(domain, data, range(10))
        np.testing.assert_equal(data1.X, data2.X)

    def test_single_call(self):
        obj = DummyPlus(Mock(return_value=1))
        self.assertEqual(obj.compute_shared.call_count, 0)
        data = Orange.data.Table("iris")[45:55]  # two classes
        domain = Orange.data.Domain([at.copy(compute_value=obj)
                                     for at in data.domain.attributes],
                                    data.domain.class_vars)

        Orange.data.Table.from_table(domain, data)
        self.assertEqual(obj.compute_shared.call_count, 1)
        ndata = Orange.data.Table.from_table(domain, data)
        self.assertEqual(obj.compute_shared.call_count, 2)

        #the learner performs imputation
        c = Orange.classification.LogisticRegressionLearner()(ndata)
        self.assertEqual(obj.compute_shared.call_count, 2)
        c(data) #the new data should be converted with one call
        self.assertEqual(obj.compute_shared.call_count, 3)

        #test with descendants of table
        DummyTable.from_table(c.domain, data)
        self.assertEqual(obj.compute_shared.call_count, 4)

    def test_compute_shared_eq_warning(self):
        with warnings.catch_warnings(record=True) as warns:
            DummyPlus(compute_shared=lambda *_: 42)

            class Valid:
                def __eq__(self, other):
                    pass

                def __hash__(self):
                    pass

            DummyPlus(compute_shared=Valid())
            self.assertEqual(warns, [])

            class Invalid:
                pass

            DummyPlus(compute_shared=Invalid())
            self.assertNotEqual(warns, [])

        with warnings.catch_warnings(record=True) as warns:

            class MissingHash:
                def __eq__(self, other):
                    pass

            DummyPlus(compute_shared=MissingHash())
            self.assertNotEqual(warns, [])

        with warnings.catch_warnings(record=True) as warns:

            class MissingEq:
                def __hash__(self):
                    pass

            DummyPlus(compute_shared=MissingEq())
            self.assertNotEqual(warns, [])

        with warnings.catch_warnings(record=True) as warns:

            class Subclass(Valid):
                pass

            DummyPlus(compute_shared=Subclass())
            self.assertNotEqual(warns, [])

    def test_eq_hash(self):
        x = Orange.data.ContinuousVariable("x")
        y = Orange.data.ContinuousVariable("y")
        x2 = Orange.data.ContinuousVariable("x")
        assert x == x2
        assert hash(x) == hash(x2)
        assert x != y
        assert hash(x) != hash(y)

        c1 = SharedComputeValue(abs, x)
        c2 = SharedComputeValue(abs, x2)

        d = SharedComputeValue(abs, y)
        e = SharedComputeValue(len, x)

        self.assertNotEqual(c1, None)

        self.assertEqual(c1, c2)
        self.assertEqual(hash(c1), hash(c2))

        self.assertNotEqual(c1, d)
        self.assertNotEqual(hash(c1), hash(d))

        self.assertNotEqual(c1, e)
        self.assertNotEqual(hash(c1), hash(e))



import unittest
from Orange.widgets.utils.localization import pl


class TestEn(unittest.TestCase):
    def test_pl(self):
        self.assertEqual(pl(0, "cat"), "cats")
        self.assertEqual(pl(1, "cat"), "cat")
        self.assertEqual(pl(2, "cat"), "cats")
        self.assertEqual(pl(100, "cat"), "cats")
        self.assertEqual(pl(101, "cat"), "cats")

        self.assertEqual(pl(0, "cat|cats"), "cats")
        self.assertEqual(pl(1, "cat|cats"), "cat")
        self.assertEqual(pl(2, "cat|cats"), "cats")
        self.assertEqual(pl(100, "cat|cats"), "cats")
        self.assertEqual(pl(101, "cat|cats"), "cats")


if __name__ == "__main__":
    unittest.main()
